#!/bin/bash
#
# Component: commit
#
# Purpose:
#   Sends Storage Commitment Response to DICOM device.
#
# Usage:
#    sendResponse.sh -d <device> -f <response file>
#
#     -d <device>        : ID of the device to sand the commitment
#     -f <response file> : Commitment Response to be sent
#
#

. $HOME/etc/setenv.sh

# if sendResponse.sh is called directly:
# run as a task instead
if [ `basename $0` = "sendResponse.sh" ] ; then
	$TASKD_HOME/runshellcmd.sh -R -r 0 $0 "$@"
	exit $?
fi


LOGGER="/usr/bin/logger"
TAG="sendCommitment[$$]"

usage_()
{
[ -n "$1" ] && echo "$1" >&2
cat << END_OF_USAGE >&2
Usage:

   sendCommitment.sh -d <device> -f <response file>

     -d <device>        : ID of the device to sand the commitment
     -f <response file> : Commitment Response to be sent

END_OF_USAGE
} # usage_

# Parameters
ResponseFile=""
HostID=""

log_()
{
	if [ -n "$1" ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "ERROR OBJECT fwd $LogUser $File $DB_STYIUID $DB_PNAME $DB_MODALITY $HostID - - - - - - $1"
	else
		$LOGGER -p $LOG_FACILITY.info -t "$TAG" "INFO OBJECT fwd $LogUser $File $DB_STYIUID $DB_PNAME $DB_MODALITY $HostID - - - - - - -"
	fi
} #log_


ERRMSG=""

while getopts 'd:f:' OPT; do
	case $OPT in
	d) HostID="$OPTARG" ;;
	f) ResponseFile="$OPTARG" ;;
	*) ERRMSG="Invalid Parameter: $OPT" ;;
	esac
done

[ -z "$HostID" ] && ERRMSG="Missing Parameter: HostID"
[ -z "$ResponseFile" ] && ERRMSG="Missing Parameter: ResponseFile"

# include nettask.shi for checking if the device is temporary down
. $TASKD_HOME/nettask.shi

# override getPeer_() of nettask.shi
getPeer_()
{
	echo "$HostID"
}


run_() # override run_ method
{
	if [ -n "$ERRMSG" ] ; then
		usage_
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "$ERRMSG"
		return 125
	fi

	if [ ! -f "$ResponseFile" ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "Response File does not exist: $ResponseFile"
		return 125
	fi

	# Check if the target device is temporarily down and return
	checkPeer_
	if [ $? -ne 0 ] ; then
		$LOGGER -p $LOG_FACILITY.error -t "$TAG" "$HostID is temporarily down - retry"
		return 1
	fi


	QUERY="select HOST,PORT,TYPE,AE,PROTOCOL from $TARGETTBL where ID='$HostID'"
	SQL_RES=$( ${TOOLS_HOME}/sql.sh "$QUERY;" -E 2>/dev/null )

	if [ $? -ne 0 ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "Error connecting to MySQL"
		WATSON_ERRMSG="Error connecting to MySQL"
		return 1
	fi

	# This line below will create DEV_<DNAME>="<VALUE>" equations and
	# evaluate them, so eg.: $DEV_HOST and DEV_AE can be used in later codes
	eval `echo "$SQL_RES" | grep -v "^\*\*" | sed -e 's/ */DEV_/' -e 's/: /="/' -e 's/$/"/'`

	$LOGGER -p $LOG_FACILITY.debug -t "$TAG" "Device: [Host: $FWD_HOST, Port: $FWD_PORT, Type: $FWD_TYPE, AE: $FWD_AE, Protocol: $FWD_PROTOCOL]"

	if [ -z "$DEV_HOST" -o -z "$DEV_PORT" -o -z "$DEV_TYPE" -o -z "$DEV_AE" -o -z "$DEV_PROTOCOL" ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "Bad entry in Device table for $HostID"
		WATSON_ERRMSG="Bad entry in Device table for $HostID"
		return 125
	fi

	. $DICOM_VAR/pb-scp.cfg

	[ -z "$AE_TITLE" ] && AE_TITLE="$DICOM_AE_TITLE"

	SCU_MYAET=""
	[ -n "$AE_TITLE" ] && SCU_MYAET="-aet $AE_TITLE"

	CALL_AE="--call $DEV_AE"

	TLS_OPT=""
	[ "$FWD_PROTOCOL" = "S" ] && TLS_OPT="+tls $SSL_SITE_KEY $SSL_SITE_CERT -ic +ps"

	if [ "$FWD_PROTOCOL" = "S" ] ; then
		$OPENSSL_PASS
	fi | \

	$DICOM_HOME/bin/sendCommitEvent $TLS_OPT $SCU_MYAET $CALL_AE --file $ResponseFile $DEV_HOST $DEV_PORT > /dev/null 2>&1

	if [ $? -ne 0 ] ; then
		$LOGGER -p $LOG_FACILITY.err -t "$TAG" "Send Storage Commitment failed $HostID [$ResponseFile]"
		WATSON_ERRMSG="Send Storage Commitment failed $DEV_HOST:$DEV_PORT:$DEV_AE"
		return 1
	fi

	$LOGGER -p $LOG_FACILITY.info -t "$TAG" "Storage Commitment sent to $HostID [$ResponseFile]"

	return 0

} # run_()


# override cleanup_
cleanup_()
{
	[ -f "$ResponseFile" ] && rm -f "$ResponseFile"
}


# override print_
print_()
{
	PrintMsg="Sending Storage Commitment Response to $HostID [$ResponseFile]"
	echo "$PrintMsg"
}


get_watson_message_()
{
	[ -z "$DrWatsonID" ] && DrWatsonID="sendCommit.$HostID"
	[ -z "$DrWatsonMessage" ] && DrWatsonMessage="Sending Storage Commitment to $HostID ($WATSON_ERRMSG)"

	if [ "$RUN_RESULT" -gt 124 ] ; then
		WATSON_MESSAGE_ID="$DrWatsonID.failed"
		WATSON_MESSAGE_TEXT="FAILED: $DrWatsonMessage"
		return
	fi

	if [ "$RUN_RESULT" -ne 0 -a "$RETRY_NUM" -gt 0 -a $(( $RETRY_NUM % 10)) -eq 0 ] ; then
		WATSON_MESSAGE_TEXT="ERROR: $DrWatsonMessage [Retry #$RETRY_NUM]"
		WATSON_MESSAGE_ID="$DrWatsonID.$RETRY_NUM"
	fi
} # get_watson_message_()

