#!/bin/bash
#
## Component: dicom
##
## Description:
##	DICOM objects comparison or md5 check sum calculation
##
##	Exit code is same as diff and cmp unix commands:
##		0 means Canonized objects are same
##    1 means Canonized objects are differ
##    2 means Error occured during checking
##
##	md5 check sum passed through standard output (if exit code is 0)
##
## Usage:
##	objdiff.sh -d [-i <ignoring_patterns_file>] [-c <compression_mode>] <dcmFile1> <dcmFile2> or
##  objdiff.sh -m [-i <ignoring_patterns_file>] <dcmFile1>              (only for computing md5 check sum)
##
##	where compression_mode is "lossy" or "lossless", default is "lossless"
##
## $Author$
## $Date$
##
#  RCS File Name    : $RCSfile$
#  RCS Revision     : $Revision$
#  Locked by        : $Locker$
#  Symbolic name    : $Name$
#  State            : $State$
#
. $HOME/etc/setenv.sh

#--- Initial parameter processing ---

usage=`cat <<-ENDofUSAGE
	USAGE: objdiff.sh -d [-i <ignoring_patterns_file>] [-c <compression_mode>] <dcmFile1> <dcmFile2> or
	       objdiff.sh -m [-i <ignoring_patterns_file>] <dcmFile1>              (only for computing md5 check sum)
ENDofUSAGE`

if [ "--help" = "$1" ] ; then
	echo "$usage"
	exit 0
elif [ "-d" = "$1" ] ; then
	MODE="diff"
elif [ "-m" = "$1" ] ; then
	MODE="md5"
else
	echo "Invalid or missed parameter: $1" >&2
	echo "$usage" >&2
	exit 2
fi

IGNORING_PATTERNS=""
if [ "-i" = "$2" ] ; then
	IGNORING_PATTERNS="$3"
	if [ ! -f "$IGNORING_PATTERNS" ] ; then
		echo "ERROR: <ignoring_patterns_file> parameter missed or $IGNORING_PATTERNS file does not exists." >&2
		echo "$usage" >&2
		exit 2
	fi
	IGNORING_PATTERNS2="-i $IGNORING_PATTERNS"
	shift 2
fi

compMode="lossless"
if [ "$MODE" = "diff" ] ; then
    if [ "-c" = "$2" ] ; then
	compMode="$3"
	shift 2
    fi
fi

DCMF1="$2"
if [ -z "$DCMF1" -o ! -f "$DCMF1" ] ; then
	echo "<dcmFile1> parameter missed or $DCMF1 file does not exists." >&2
	echo "$usage" >&2
	exit 2
fi
if [ "$MODE" = "diff" ] ; then
	DCMF2="$3"
	if [ -z "$DCMF2" -o ! -f "$DCMF2" ] ; then
		echo "<dcmFile2> parameter missed or $DCMF2 file does not exists." >&2
		echo "$usage" >&2
		exit 2
	fi
fi

#--- Compare shortcut ---

if [ "$MODE" = "diff" ] ; then
	cmp -s $DCMF1 $DCMF2
	[ $? -eq 0 ] && exit 0
fi

DATE=`date +"%Y%m%d%H%M%S"`
NORMFORM1="$TMP/objnorm1.$DATE.$$"
NORMFORM2="$TMP/objnorm2.$DATE.$$"
ERRMSG="$TMP/objnormerr.$DATE.$$"

$DICOM_HOME/bin/objcanonize.sh $IGNORING_PATTERNS2 $DCMF1 >$NORMFORM1 2>$ERRMSG
if [ $? -ne 0 ] ; then
	echo "ERROR: during canonizing $DCMF1:" >&2
	cat $ERRMSG >&2
	rm -f $NORMFORM1 $ERRMSG
	exit 2
fi

#--- md5 computing ---

if [ "$MODE" = "md5" ] ; then
	cat $NORMFORM1 | md5sum -b
	RES=$?
	rm -f $NORMFORM1 $ERRMSG
	exit $RES
fi

$DICOM_HOME/bin/objcanonize.sh $IGNORING_PATTERNS2 $DCMF2 >$NORMFORM2 2>$ERRMSG
if [ $? -ne 0 ] ; then
	echo "ERROR: during canonizing $DCMF2:" >&2
	cat $ERRMSG >&2
	rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
	exit 2
fi

#--- Compare canonized ---
if [ "$compMode" = "lossy" ] ; then
    DATE=`date +"%Y%m%d%H%M%S"`
    # preparing work_dir1
    WORK_DIR1="$TMP/objnorm.$DATE.$$.1"
    mkdir $WORK_DIR1
    if [ ! -d $WORK_DIR1 ] ; then
	echo "ERROR: working directory can not be created." >&2
	rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
	exit 16
    fi
    ERR=`$DCVIEWER_HOME/herelo -v -e -s $DCMF1 $WORK_DIR1/obj 2>&1 >$WORK_DIR1/dump`
    if [ $? -ne 0 ] ; then
        echo "ERROR: can not herelo $DCMF1:" >&2
        echo "$ERR" >&2
        rm -rf $WORK_DIR1
	rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
        exit 125
    fi
    if [ -n "$IGNORING_PATTERNS" ] ; then 
	cat $WORK_DIR1/dump | grep -vf "$IGNORING_PATTERNS" >$WORK_DIR1/dump.tmp
	rm $WORK_DIR1/dump
	mv $WORK_DIR1/dump.tmp $WORK_DIR1/dump
    fi
    cat $WORK_DIR1/dump | sed 's;^\(([0-9a-f,]*)\) ?? ;\1 UN ;' >$WORK_DIR1/obj
    rm -f $WORK_DIR1/dump
    rm -f $WORK_DIR1/obj.desc
    for i in `ls "$WORK_DIR1"|grep "128x128"` ; do 
	rm -f $WORK_DIR1/$i ; 
    done
    # preparing work_dir2
    WORK_DIR2="$TMP/objnorm.$DATE.$$.2"
    mkdir $WORK_DIR2
    if [ ! -d $WORK_DIR2 ] ; then
	echo "ERROR: working directory can not be created." >&2
	rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
	rm -rf $WORK_DIR1
	exit 16
    fi
    ERR=`$DCVIEWER_HOME/herelo -v -e -s $DCMF2 $WORK_DIR2/obj 2>&1 >$WORK_DIR2/dump`
    if [ $? -ne 0 ] ; then
        echo "ERROR: can not herelo $DCMF2:" >&2
        echo "$ERR" >&2
        rm -rf $WORK_DIR2 $WORK_DIR1
	rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
        exit 125
    fi
    if [ -n "$IGNORING_PATTERNS" ] ; then 
	cat $WORK_DIR2/dump | grep -vf "$IGNORING_PATTERNS" >$WORK_DIR2/dump.tmp
	rm $WORK_DIR2/dump
	mv $WORK_DIR2/dump.tmp $WORK_DIR2/dump
    fi
    cat $WORK_DIR2/dump | sed 's;^\(([0-9a-f,]*)\) ?? ;\1 UN ;' >$WORK_DIR2/obj
    bitsallocated=`cat $WORK_DIR2/dump | grep "(0028,0100)"|awk '{ print $3}'`
    rm -f $WORK_DIR2/dump
    rm -f $WORK_DIR2/obj.desc
    for i in `ls "$WORK_DIR2"|grep "128x128"` ; do 
	rm -f $WORK_DIR2/$i ; 
    done
    # comparing objects
    OBJ1="${WORK_DIR1}/obj"
    OBJ2="${WORK_DIR2}/obj"
    cmp -s $OBJ1 $OBJ2
    RES=$?
    if [ $RES -eq 0 ] ; then
	rm -f $OBJ1
	rm -f $OBJ2
	DIR1=`ls "${WORK_DIR1}"|sort`
	DIR2=`ls "${WORK_DIR2}"|sort`
	if [ "$DIR1" != "$DIR2" ] ; then
	    rm -rf $WORK_DIR2
	    rm -rf $WORK_DIR1
	    rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
	    echo "$DCMF1 and $DCMF2 are different."
	    exit 1
	fi
	MSEARG="-b $bitsallocated"
	for image in `ls "${WORK_DIR1}"|sort` ; do 
	    echo "comparing $image"
	    $UTILS_HOME/bin/MSECompare -f "${WORK_DIR1}/$image" -c "${WORK_DIR2}/$image" "$MSEARG" -v
	    RES=$?
	    if [ $RES -ne 0 ] ; then 
		[ $RES -eq 1 ] && echo "$DCMF1 and $DCMF2 are different."
    	        rm -rf $WORK_DIR2
		rm -rf $WORK_DIR1
		rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
		exit $RES
	    fi
	done
    fi
    
    rm -rf $WORK_DIR2
    rm -rf $WORK_DIR1
else
    cmp -s $NORMFORM1 $NORMFORM2
    RES=$?
fi
rm -f $NORMFORM1 $NORMFORM2 $ERRMSG
[ $RES -eq 1 ] && echo "$DCMF1 and $DCMF2 are different."
exit $RES
