#ifndef __DICOMPARSER_H_INCLUDED__
#define __DICOMPARSER_H_INCLUDED__


#include "dcmtk/ofstd/oftypes.h"
#include "dcmtk/ofstd/ofstring.h"

#include "pbdcm/PBItem.h"
#include "pbdcm/defpbelems.h"

class IDicomTagParser2;

/**
 * DICOM tag data.
 * Storing tag and range.
 */
class IDicomTagParserData
{
public:
	unsigned int 			sTag;
	DcmTagKey				sDcmTag;
	unsigned int			sEndTag; // normally sTag + 1

	IDicomTagParserData		*pNext;


	IDicomTagParserData();
	virtual ~IDicomTagParserData() {}

	virtual bool CheckTag( unsigned int eTag, IDicomTagParser2 *obj, void *pData) = 0;
	virtual void *InitTag() = 0;
	virtual void ReleaseTagData( void *) = 0;
};

/**
 * DICOM one tag data.
 * Processing function.
 */
class SDicomTagParserData2 : public IDicomTagParserData
{
public:
	typedef		void (processTagFunc)( IDicomTagParser2 *obj);

	processTagFunc			*pFunc;

public:

	SDicomTagParserData2();

	bool CheckTag( unsigned int eTag, IDicomTagParser2 *obj, void * /*pData*/) /*override*/ {
		if (sTag <= eTag) {
			if (sEndTag > eTag) {
				pFunc( obj);
				return (sEndTag != (eTag + 1));
			} else {
				// over range, no more test
				return false;
			}
		}
		return true;
	}

	void *InitTag() /*override*/ {return NULL;}
	void ReleaseTagData( void *) /*override*/ {}
};




#ifdef RANGEINTERSECTION


/**
 * Linked list container for current processed IDicomTagParserData.
 * This list containes that DICOM tag data, which range is in current DICOM element.
 */
class IDicomTagParserDataList
{
public:
	IDicomTagParserDataList		*pNext;
	IDicomTagParserData			*pData;
	void						*pGroupData;
};

class SDicomTagGroupList;

/**
 * Linked list of IDicomTagParserDataList container.
 * .
 */
class SDicomTagParserDataListContainer
{
	friend class SDicomTagGroupList;
protected:
	/** maximum intersected DICOM tag data */
	unsigned int				uiMaxintersectedRegion;
	/** Number of DICOM tag data */
	unsigned int				uiProcessedTagNumber;

	// dynamic
	/** Linked list of current used IDicomTagParserDataList. First element is a dummy one, which has NULL data, and the last one of the list is hooked to it. */
	IDicomTagParserDataList		pCurrDataList;
	/** Allocated but not used IDicomTagParserDataList. First element is a dummy one, which has NULL data, and the last one of the list is hooked to it.  */
	IDicomTagParserDataList		pFreeDataList;

	/** Allocating the pFreeDataList. */
	void						initIntersectedRegions();
	/** Releasing the allocated resources. */
	void						releaseIntersectedRegions();

protected:
	/** Get a IDicomTagParserDataList from the pFreeDataList and add to the pCurrDataList. */
	IDicomTagParserDataList *	addOneToCurrDataList();
	/** Removing the IDicomTagParserDataList from the pCurrDataList and add to the pFreeDataList. */
	void						removeFromCurrDataList( IDicomTagParserDataList *&);

	/** Calculating the uiMaxintersectedRegion. */
	void						analyzeFunctions( IDicomTagParserData *pFirst);

public:
	SDicomTagParserDataListContainer() {
		uiMaxintersectedRegion	= 0;
		uiProcessedTagNumber	= 0;

		pCurrDataList.pNext		= &pCurrDataList;
		pCurrDataList.pData		= NULL;
		pFreeDataList.pNext		= &pCurrDataList;
		pFreeDataList.pData		= NULL;
	}

	virtual ~SDicomTagParserDataListContainer() {
		releaseIntersectedRegions();
	}


};
#endif

//----------------------------------
class SDicomTagParserGroupDataDynData
{
public:

	IDicomTagParserData			*pNext;
	IDicomTagParserData			*pCurrent;

	unsigned int		ntag;
	unsigned int		netag;

	unsigned int		lastGroupID;

	SDicomTagParserGroupDataDynData() {
		lastGroupID = 0;
		ntag        = 0;
		netag       = 0;
		pNext       = NULL;
		pCurrent    = NULL;
	}
	virtual ~SDicomTagParserGroupDataDynData() {}
};


//----------------------------------
class SDicomTagParserGroupData : public IDicomTagParserData
#ifdef RANGEINTERSECTION
									, public SDicomTagParserDataListContainer
#endif
{
protected:

	IDicomTagParserData *		pFirst;

public:
	SDicomTagParserGroupData() {
		pFirst			= NULL;
	}

	~SDicomTagParserGroupData() {
		removeAllTags();
	}

	void *InitTag() /*override*/ {

		return new SDicomTagParserGroupDataDynData;
	}
	void ReleaseTagData( void *pData) /*override*/ {
		delete ((SDicomTagParserGroupDataDynData *)pData);
	}

	void ResetTag( SDicomTagParserGroupDataDynData* pData) {
		pData->pNext	= pFirst;
		pData->ntag		= pFirst->sTag;
		pData->netag	= pFirst->sEndTag;
		pData->pCurrent	= NULL;
	}

	virtual bool FilteredTag( unsigned int eTag);

	bool CheckTag( unsigned int eTag, IDicomTagParser2 *obj, void *pData) /*override*/ ;

	void addTag( unsigned int sTag, const DcmTagKey &tag, SDicomTagParserData2::processTagFunc *func, const char *privateCreator, unsigned int length);
	void removeAllTags();

#ifdef RANGEINTERSECTION
	void analyzeFunctions() 
	{ 
		SDicomTagParserDataListContainer::analyzeFunctions( pFirst);
		initIntersectedRegions();
	}
#endif
};


//----------------------------------
class SDicomTagParserPrivateGroupDataDynData : public SDicomTagParserGroupDataDynData
{
public:
	unsigned int		privateBlockID;
};

//----------------------------------
class SDicomTagParserPrivateGroupData : public SDicomTagParserGroupData
{
public:

	const char *privateCreator;

public:

	void *InitTag() /*override*/ {
		return new SDicomTagParserPrivateGroupDataDynData;
	}
	void ReleaseTagData( void *pData) /*override*/ {
		delete ((SDicomTagParserPrivateGroupDataDynData *)pData);
	}

	inline void ResetPrivateTag( SDicomTagParserPrivateGroupDataDynData* pData) {

		SDicomTagParserGroupData::ResetTag( pData);

		pData->privateBlockID	= -1;
	}

	bool FilteredTag( unsigned int eTag) /*override*/ ;

	bool CheckTag( unsigned int eTag, IDicomTagParser2 *obj, void *pData) /*override*/ ;
};





//----------------------------------
class SDicomTagGroupList
{
public:
	IDicomTagParserData	*pFirst;

public:
	SDicomTagGroupList();
	virtual ~SDicomTagGroupList()
	{
		ReleaseTags();
	}


	// init functions
	static unsigned int	getTagID( const DcmTagKey &_tag)
		{
			unsigned int etag	= _tag.getElement();
			unsigned int gtag	= _tag.getGroup();
			return ((gtag << 16) | (etag & (((((gtag & 1) ^ 1) << 8) + 1) * 0xff)));
		}
	static unsigned int	getRawTagID( const DcmTagKey &_tag)
		{
			unsigned int etag	= _tag.getElement();
			unsigned int gtag	= _tag.getGroup();
			return ((gtag << 16) | (etag));
		}
	void				addTag( const DcmTagKey &tag, SDicomTagParserData2::processTagFunc *func, const char *privateCreator = NULL, unsigned int length = 0x10001);
	virtual void		initFunction();

	// release functions
	void	ReleaseTags();

#ifdef RANGEINTERSECTION
	void			analyzeFunctions();
#endif
};


//----------------------------------
class IDicomTagParser2
{
protected:
	bool			bRemoveElem;

public:
	DcmElement *	elem;
	void *			owner;

public:
	virtual void	removeElement() = 0;

	IDicomTagParser2()
	{
		elem         = NULL;
		bRemoveElem  = false;
		owner        = NULL;
	}
	virtual ~IDicomTagParser2() {}
};


bool NoPreprocessFunction( IDicomTagParser2 *obj);

//----------------------------------
class CDicomTagParser2 : public IDicomTagParser2
#ifdef RANGEINTERSECTION
									, public SDicomTagParserDataListContainer
#endif
{
public:
	typedef		bool (preprocessFunc)( IDicomTagParser2 *obj);
protected:

	DcmItem*					item;
	unsigned int				elemID;

	preprocessFunc				*fPreprocessFunction;

	IDicomTagParserData			*pFirst;

protected:

	inline DcmElement *		getNextDcmElement( DcmItem* _item, void *_owner, DcmElement *);
	virtual void			InitPrivateData() = 0;

public:

	CDicomTagParser2() : IDicomTagParser2() {
		elemID              = 0;
		item                = NULL;
		fPreprocessFunction = NoPreprocessFunction;
		pFirst              = NULL;
	}

	virtual ~CDicomTagParser2() {}

	// parse
	void	parseTags( DcmItem* _item, void *owner, SDicomTagGroupList *groupList, preprocessFunc *func = NoPreprocessFunction);

	// in parse
	void	removeElement();

};




#endif // __DICOMPARSER_H_INCLUDED__
