/**
 * Component: imgproc
 *
 * Description:
 *    Abstract base class for data extracted from DICOM objects. There are two known subclasses:
 *       - CRawImage for extracted Image Frames
 *       - CDocument for extracted Documents (pdf)
 *       - CPbRContent for extracting obsolescence information from PbR files
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */

#ifndef CEXTRACTEDDATA_H_
#define CEXTRACTEDDATA_H_

#ifndef IMGPROC_NO_DCMTK
#include "dcmtk/ofstd/oftypes.h"
#else
#include "imgproc/nodcmtk.h"
#endif //IMGPROC_NO_DCMTK

class CImageProcess;

/**
 * Abstract base class for data extracted from DICOM objects. There are two known direct subclasses:
 *       - CRawImage for extracted Image Frames
 *       - CDocument for extracted Documents (pdf)
 *       - CPbRContent for extracting obsolescence information from PbR files
 */
class CExtractedData {
	protected:

		/** Image Process object holding original DICOM data */
		CImageProcess * m_imageProcess;

		/** Extracted content */
		char * m_contentData;

		/** Length of extracted content */
		Uint32 m_contentLength;

		/** Index of extracted data from the DICOM object */
		int m_index;

	public:

		/**
		 * Constructor for creating new CExtractedData object.
		 *
		 * @param imageProcess   Image Process object holding original DICOM data
		 * @param index          Extracted Data index within the DICOM object
		 */
		CExtractedData( CImageProcess * imageProcess, int index );

		/**
		 * Destructor for destroying CExtractedData object.
		 */
		virtual ~CExtractedData();

		/**
		 * Initializes CExtractedData object for processing.
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool init() = 0;

		/**
		 * Post processes extracted data.
		 *
		 * @return 0 on success; >0 on error
		 */
		virtual int postProcess() = 0;

		/**
		 * Retrieves Extracted Data index within the DICOM object.
		 * (Extracted Data index is image/frame index for image objects)
		 *
		 * @return Extracted Data index within the DICOM object
		 */
		inline int getIndex() const {
			return m_index;
		}

};

#endif /* CEXTRACTEDDATA_H_ */
