/*
 * Component:  common
 *
 * Purpose:
 *	    Header file for DICOM image object processing
 *
 * Last Updated     : $Author: tnemeth $
 * Update Date Time : $Date: 2013/12/04 13:48:51 $
 * RCS File Name    : $RCSfile: imgproc/imageprocess.h,v $
 * RCS Revision     : $Revision: 1.9 $
 * Locked by        : $Locker:  $
 * Symbolic name    : $Name:  $
 * State            : $State: Exp $
 */
#ifndef __IMAGE_PROCESS_H_INCLUDED__
#define __IMAGE_PROCESS_H_INCLUDED__

#ifndef IMGPROC_NO_DCMTK
#include "dcmtk/ofstd/oftypes.h"
#include "dcmtk/ofstd/ofstring.h"
#else
#include "imgproc/nodcmtk.h"
#endif //IMGPROC_NO_DCMTK
#include "imgproc/dicomdefine.h"
#include "imgproc/overlay.h"
#include "imgproc/histogram.h"
#include "imgproc/imagememory.h"
#include "imgproc/rawimage.h"
#ifndef IMGPROC_NO_DCMTK
#include "imgproc/encapsulateddocument.h"
#include "imgproc/pbrcontent.h"
#include "imgproc/dicomparser.h"
#endif //IMGPROC_NO_DCMTK

class IImageProcessProgress
{
public:
	virtual ~IImageProcessProgress() { }
	virtual void onImageConverted() = 0;
	virtual void onFrameWritten() = 0;
};

/**
 * DICOM image object processing.
 * Processes dicom dataset by getting overlay and raw pixel data.
 */
class CImageProcess
#ifndef IMGPROC_NO_DCMTK
		: protected CDicomTagParser2, public SDicomTagGroupList
#endif //IMGPROC_NO_DCMTK
{
	private:
		/** DICOM dataset to be processed */
		DcmDataset * m_dataset;

		// Cummulated and array values

		/** Array for Actual Bits Used (dada) */
		Uint16 * m_actualBitsUsedArray;
		/** Size of the Actual Bits Used Array */
		int m_actualBitsUsedArraySize;
		/** Current element in the Actual Bits Used Array */
		int m_actualBitsUsedArrayCount;

		/** PB Min Value (ba21) Array */
		Sint32 * m_minValueArray;
		/** Size of the PB Min Value Array */
		int m_minValueArraySize;
		/** Current element in the PB Min Value Array */
		int m_minValueArrayCount;

		/** PB Max Value (ba22) Array */
		Sint32 * m_maxValueArray;
		/** Size of the PB Max Value Array */
		int m_maxValueArraySize;
		/** Current element in the PB Max Value Array */
		int m_maxValueArrayCount;

		/** Best Window (ba11) Array */
		Sint32 * m_bestWindowArray;
		/** Best Window 2 (ba11) Array */
		Sint32 * m_bestWindow2Array;
		/** Size of the Best Window Array */
		int m_bestWindowArraySize;
		/** Current element in the Best Window Array */
		int m_bestWindowArrayCount;

		/** Best Level (ba12) Array */
		Sint32 * m_bestLevelArray;
		/** Best Level 2 (ba12) Array */
		Sint32 * m_bestLevel2Array;
		/** Size of the Best Level Array */
		int m_bestLevelArraySize;
		/** Current element in the Best Level Array */
		int m_bestLevelArrayCount;


		// Members created due to refactoring

		// indicates whether the object contains pixel data or not
		OFBool m_hasImage;
		// exit code for indicating success or failure
		int m_exitCode;

public:
#ifndef IMGPROC_NO_DCMTK
		virtual void initFunction() /* override*/ ;
#endif //IMGPROC_NO_DCMTK

	protected:

		/** Output stream */
		std::ostream* m_pOut;

		// Members created due to refactoring

		void incrementExitCode( int inc ) {
			m_exitCode += inc;
		}

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Reading DICOM tags defined in tag registry from a DICOM item/dataset
		 *
		 * @param item       DICOM item/dataset to read the DICOM tags from
		 * @param registry   registry of DICOM tags to be read from the DICOM item
		 *
		 * Defined by CDicomTagParser2
		 */
		inline void parseTags( DcmItem * item /*, preprocessFunc *func = NoPreprocessFunction*/) {
			CDicomTagParser2::parseTags( item, NULL, this, fTagParserPreprocess /* func*/ );
		}
#endif //IMGPROC_NO_DCMTK

		bool checkAndSetCurrentOverLay();


		/**
		 * Initializes CDicomTagParser2 private data
		 *
		 * Defined by CDicomTagParser2
		 */
		virtual void InitPrivateData();


		/** Image memory manager used to get memory for raw pixel data */
		CImageMemory * m_imageMemory;

		/** Columns in the image(s) stored in the DICOM object */
		Uint16 m_columns;

		/** Rows in the image(s) stored in the DICOM object */
		Uint16 m_rows;

		/** Bits Allocated */
		Uint16 m_bitsAllocated;

		/** Bits Stored */
		Uint16 m_bitsStored;

		/** High Bit */
		Uint16 m_highBit;

		/** Pixel Representation */
		Uint16 m_pixelRepresentation;

		/** Planar Configuration */
		Uint16 m_planarConfiguration;

		/** Samples Per Pixel */
		Uint16 m_samplesPerPixel;
		/** Set Samples Per Pixel */
		void setSamplesPerPixel( Uint16 samplesPerPixel, DcmElement* elem );
		DcmElement* m_pSamplesPerPixelElement;

		/** Number of Frames/Slices */
		Uint16 m_frameNumber;

		/** Rescale Intercept */
		double m_rescaleIntercept;

		/** Rescale Slope */
		double m_rescaleSlope;

		/** Window Center */
		double m_windowCenter;

		/** Window Width */
		double m_windowWidth;

		/** Pixel Spacing X */
		double m_pixelSpacingX;
		/** Pixel Spacing Y */
		double m_pixelSpacingY;

		/** Modality */
		OFString m_modality;

		/** Photometric Interpretation */
		PI_TYPE m_photometricInterpretation;
		/** Set photometric implementation from string */
		void setPhotometricInterpretation( OFString& colorModel, DcmElement* elem );
		DcmElement* m_pPhotometricInterpretationElement;

		/** Have planar configuration */
		bool m_bHavePlanarConfig;

		// For checking SuperDASM 1.0 (sends false Photometric Interpretation)
		/** Secondary Capture Device Model Name */
		OFString m_SCDModelName;
		/** Secondary Capture Device Software Version */
		OFString m_SCDSoftwareVersion;
		/** For inverting images received with false Photometric Interpretation from SuperDASM 1.0 */
		OFBool m_isInverse;

		// Palette Color Lookup Table Data
		/** Lookup table data for Red, Blue and Green components */
		Uint16 * m_paletteData[ 3 ];
		/** Length of palette */
		Uint16 m_paletteLength[ 3 ];
		/** sample size in the palette */
		Uint16 m_paletteSampleSize[ 3 ];

		/** Image Type */
		ImageType m_imageType;

		// Overlays

		/** Overlay array */
		COverlayArray m_overlayArray;
		/** Index of current overlay */
		int m_currentOverlay;
		/** Group of current overlay */
		int m_currentOverlayGroup;
		/** Bit mask for getting overlay data from pixel data */
		Uint16 m_overlayMask;

		/** Pixel Data content */
		Uint8 * m_pixelData;

		/** Pixel Data length */
		Uint32 m_pixelDataLength;

		/** Number of pixels in the image */
		Uint32 m_pixelPerImage;

		/** Number of pixels to be processed per image
		 * generally = m_pixelPerImage, except in case of YBR_*_422 */
		Uint32 m_inputPixelPerImage;

		/** Raw image processing (frame) */
		CRawImage * m_rawImage;

		/** Histogram for calculating best window/level */
		CHistogram m_histogram;

		/** Current image */
		int m_imageCounter;

		/** Specific frame to Process (not zero-based, first image is '1')
		 * if non-zero, ONLY the specific frame will be processed and written,
		 * desc and processed dicom WILL NOT be written,
		 * (compression flags and thumb setting are respected) */
		int m_specificFrame;

		/** Also dump the content of the DICOM dataset */
		OFBool m_verbose;
		/** Dump only mode, do not process the DICOM dataset */
		OFBool m_dumponly;
		/** Escape new lines in dumps */
		OFBool m_escapeNewLines;

		/** OFTrue indicates that all 3rd party private tags should be kept
		 *  This flag is intended to be set for processing objects by the CD viewer
		 */
		OFBool m_keepAllPrivateTags;

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Checks if the Device Model Name is SuperDASM and softwareVersion is 1.0.
		 * This device sends false photometric interpretation and the image should be inverted
		 *
		 * @param modelName            Secondary Capture Device Model Name
		 * @param softwareVersion      Secondary Capture Device Software Version
		 *
		 * @return OFTrue if and only if the device is SuperDASM 1.0
		 */
		OFBool checkSCDevice( const OFString & modelName, const OFString & softwareVersion );

		/**
		 * Reads palette length and palette sample size from Color Palette Lookup Table Descriptor element
		 *
		 * @param elem   Supposed to be a pointer to a Color Palette Lookup Table Descriptor element
		 * @param color  Code of color (0 - blue; 1 - Green; 2 - Red)
		 */
		void readPaletteDesc( DcmElement * elem, int color );

		/**
		 * Reads palette data from Color Palette Lookup Table element
		 *
		 * @param elem   Supposed to be a pointer to a Color Palette Lookup Table element
		 * @param color  Code of color (0 - blue; 1 - Green; 2 - Red)
		 */
		void readPaletteData( DcmElement * elem, int color );

		/**
		 * Processes DICOM item/dataset. Collects values available for image processing and processes
		 * Pixel Data element. Will dump the item if the verbose mode is turned on.
		 *
		 * @param item    DICOM item to be processed
		 *
		 * @return 0 on success; >0 on error
		 */
		int processDataItem( DcmItem * item );
#endif //IMGPROC_NO_DCMTK
		/**
		 * Processes Pixel Data element. Read pixel data bytes (m_pixelData), and process it.
		 *
		 * @param pixelData    Pixel Data element to be processed
		 * @param evr          Value Representation of the Pixel Data element (OB/OW)
		 *
		 * @return 0 on success; >0 on error
		 */
		int processImageElem( DcmElement * pixelData,
#ifndef IMGPROC_NO_DCMTK
				const DcmEVR & evr
#else
				int pixelDataBytes
#endif //IMGPROC_NO_DCMTK
		);

		/**
		 * Processes pixel data bytes (m_pixelData).
		 *
		 * @return 0 on success; >0 on error
		 */
		int processImageBytes();

		/**
		 * If Bits Allocated is less than 8, then makes bytes from m_pixelData, and processes the
		 * result as a byte array
		 *
		 * @return 0 on success; >0 on error
		 */
		int adjustToByteBoundary();

		/**
		 * If Bits Allocated is less than 16 but greater than 8, then makes words from m_pixelData, and processes the
		 * result as a word array
		 *
		 * @return 0 on success; >0 on error
		 */
		int adjustToWordBoundary();

		/**
		 * Adds word samples to be processed into raw image data.
		 *
		 * @param block   array of word samples to be processed
		 * @param length  number of samples in the block
		 */
		void addSamples( Uint16 * block, Uint32 length );

		/**
		 * Adds byte samples to be processed into raw image data.
		 *
		 * @param block   array of byte samples to be processed
		 * @param length  number of samples in the block
		 */
		void addSamples( Uint8 * block, Uint32 length );

		/**
		 * Shifts the high bit to the 7th bit.
		 * Will be used for color images with pixel samples larger than a byte.
		 *
		 * @param block   array of word samples to be shifted
		 * @param length  number of samples in the block
		 */
		void makeBytes( Uint16 * block, Uint32 length );

		/**
		 * Shifts the high bit to the 7th bit.
		 * Will be used for images with pixel samples smaller than a byte.
		 *
		 * @param block   array of byte samples to be shifted
		 * @param length  number of samples in the block
		 */
		void makeBytes( Uint8 * block, Uint32 length );

		/**
		 * Remove overlays from pixel data. Overlays will be stored in m_overlayArray.
		 * Not used bits (when BitsStored != BitsAllocated) will be cleared.
		 *
		 * @return 0 on success; >0 on error
		 */
		int getAndRemoveOverlays();

		/**
		 * Start to process a new frame. It will initialize new image processing.
		 * Should be called before adding the first sample of the frame.
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		OFBool imageStart();

		/**
		 * Finalization of frame processing.
		 * Should be called after adding the last sample of the frame.
		 * Histogram analyzis, image post processing will be called.
		 */
		void imageDone();

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Prints the content of a DICOM item to the standard output.
		 * Will called only in verbose mode.
		 *
		 * @param item   DICOM item to be printed
		 * @param depth  Depth of DICOM element nesting (started from 0)
		 */
		void printItem( DcmItem * item, int depth );

		/**
		 * Prints the content of a DICOM sequence to the standard output.
		 * Will called only in verbose mode.
		 *
		 * @param seq    DICOM sequence to be printed
		 * @param depth  Depth of DICOM element nesting (started from 0)
		 */
		void printSequence( DcmSequenceOfItems * seq, int depth );

		/** Helper variable for print*() methods
		 * Will be used to add spaces in front of the nested elements */
		static char m_spaces[];

		/**
		 * Escapes the new lines in str. 0x0A will be encoded as \n while 0x0D will be encoded as \d.
		 *
		 * If str contains no new line character NULL will be returned.
		 * Otherwise a new string will be allocated (by new) and the escaped string will be stored in it.
		 *
		 * @param str  string to be escaped
		 *
		 * @return NULL if str contains no new line
		 *         pointer to a newly allocated escaped string (to be freed by delete[])
		 */
		char * escapeNewLines( const char * str );

		/**
		 * Prints the content of a DICOM element to the standard output.
		 * Will called only in verbose mode.
		 *
		 * @param elem   DICOM element to be printed
		 * @param depth  Depth of DICOM element nesting (started from 0)
		 */
		inline void printElement( DcmElement * elem, int depth ) {

			if ( elem->getVR() == EVR_SQ ) {
				printSequence( (DcmSequenceOfItems *) elem, depth );
			}
			else if ( m_escapeNewLines ) {
				if(!elem->valueLoaded()) {
					if( NULL != dynamic_cast<DcmByteString*>(elem) )
					    elem->loadAllDataIntoMemory();
				}
				char * f = elem->toString();
				char * e = escapeNewLines( f );
				//printf( "%s(%04x,%04x) %s %s\n", m_spaces + 32 - depth, elem->getGTag(), elem->getETag(), elem->getTag().getVRName(), e
				//		== NULL ? f : e );
				printElementSpecifier( elem->getGTag(), elem->getETag(), elem->getTag().getVRName(), depth);
				(*m_pOut) << " " << (e == NULL ? f : e) << '\n';
				if ( e != NULL ) {
					delete[] e; // e is created by new
				}
				free( f ); // f is created by malloc
			}
			else {
				if(!elem->valueLoaded()) {
					if( NULL != dynamic_cast<DcmByteString*>(elem) )
						elem->loadAllDataIntoMemory();
				}
				char * f = elem->toString();
				//printf( "%s(%04x,%04x) %s %s\n", m_spaces + 32 - depth, elem->getGTag(), elem->getETag(), elem->getTag().getVRName(), f );
				printElementSpecifier( elem->getGTag(), elem->getETag(), elem->getTag().getVRName(), depth);
				(*m_pOut) << " " << f << '\n';
				free( f );
			}
		}

		/** buffer to re-use by printElementSpecifier */
		char groupElement[12];

		/**
		 * Prints the formated specifiers of the DICOM element to the standard output.
		 * Will called only in verbose mode.
		 *
		 * @param group  	DICOM tag group
		 * @param element   DICOM tag element
		 * @param vr   		DICOM tag value representation name
		 * @param depth  	Depth of DICOM element nesting (started from 0)
		 */
		inline void printElementSpecifier( Uint16 group, Uint16 element, const char* vr, int depth ) {
			snprintf( groupElement, 12, "(%04x,%04x)", group, element );
			groupElement[11]='\0';
			(*m_pOut) << (m_spaces + 32 - depth) << groupElement << " " << vr;
		}
#endif //IMGPROC_NO_DCMTK

		/**
		 * Creates a new CRawImage object.
		 *
		 * @return pointer to the newly created CRawImage object
		 */
		virtual CRawImage * createRawImage() = 0;

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Creates a new CEncapsulatedDocument object.
		 *
		 * @param type     Document type
		 * @param length   Length of the encapsulated document in bytes
		 *
		 * @return pointer to the newly created CEncapsulatedDocument object
		 */
		virtual CEncapsulatedDocument * createEncapsulatedDocument( DOCUMENT_TYPE type, Uint32 length ) = 0;


		/**
		 * Creates a new CPbRContent object.
		 * This method will return an instance to the CPbRDocument base class.
		 * Derived classes may want to override this method to return a specialized derivation of CPbRDocument.
		 *
		 * @return pointer to the newly created CPbRDocument object
		 */
		virtual CPbRContent * createPbRContent();
#endif //IMGPROC_NO_DCMTK

		/**
		 * Post processing extracted data
		 * (eg histogram analyzis, planar, color conversions, ... for image frames)
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool postProcess( CExtractedData * data );


		/**
		 * Initialization for DICOM dataset processing.
		 * Base implementation does nothing, Herelo can
		 * register converters here
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool initProcess();

		/**
		 * Finalization of DICOM dataset processing.
		 * Reference and calculated values will be filled in the DICOM dataset
		 *
		 * @return OFTrue on success; OFFalse on error
		 */
		virtual OFBool finalizeProcess();

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Decides if the Private DICOM Tag should remain in the dataset or can be removed.
		 * Base implementation does nothing, it's for the herelo
		 *
		 * @param tag   Private DICOM Tag to be checked
		 *
		 * @return OFTrue if the tag should be kept in the DICOM dataset
		 */
		virtual OFBool isIncluded( const DcmTag & tag );
#endif //IMGPROC_NO_DCMTK

		/**
		 * Adjust word samples as words.
		 * Base implementation does nothing, it's for the viewer
		 *
		 * @param block   array of word samples to be adjusted
		 * @param length  number of samples in the block
		 */
		virtual void adjustWords( Uint16 * block, Uint32 length );

		/**
		 * Adjust word samples as bytes.
		 * Base implementation does nothing, it's for the viewer
		 *
		 * @param block   array of word samples to be adjusted
		 * @param length  number of samples in the block
		 */
		virtual void adjustBytes( Uint16 * block, Uint32 length );

		/**
		 * Adjust byte samples as bytes.
		 * Base implementation does nothing, it's for the viewer
		 *
		 * @param block   array of byte samples to be adjusted
		 * @param length  number of samples in the block
		 */
		virtual void adjustBytes( Uint8 * block, Uint32 length );

#ifndef IMGPROC_NO_DCMTK
		/**
		 * Processes Encapsulated Document element to extract document content
		 *
		 * @param elem   Encapsulated Document element taken from the dataset
		 *
		 * @return 0 on success; >0 on error
		 */
		int processEncapsulatedDocument( DcmElement * elem );

		/**
		 * Processes PbR objects to extract obsolescence information
		 *
		 * @return 0 on success; >0 on error
		 */
		int processPbRContent();
#endif //IMGPROC_NO_DCMTK

	public:

		/**
		 * Constructor for creating a new CImageProcess object.
		 *
		 * @param dataset         DICOM dataset to be processed
		 * @param imageMemory     Memory manager to get memory for raw images
		 * @param verbose         OFTrue indicates verbose mode, the dataset content will be dumped
		 * @param dumponly        OFTrue indicates to dump only, do not process the dataset
		 * @param escapeNewLines  OFTrue indicates to escape new lines in the dumps
		 * @param specificFrame   Number of specific frame to process, if non-zero,
		 *                        only image data for the specific frame will be processed and written,
		 *                        descriptor and processed dicom WILL NOT be written,
		 */
		CImageProcess( DcmDataset * dataset, CImageMemory * imageMemory, OFBool verbose = OFFalse, OFBool dumponly = OFFalse,
				OFBool escapeNewLines = OFFalse, int specificFrame = 0 );

		/**
		 * Destructor for destroying CImageProcess object.
		 */
		virtual ~CImageProcess();

		/**
		 * Processes the DICOM dataset.
		 *
		 * @return 0 on success; >0 on error
		 */
#ifndef IMGPROC_NO_DCMTK
		int process();
#else
		int process(void* pPixelData, int pixelDataBytes, int nSamplesPerPixel, const char* cstrPhotometricInterpretation, int nPlanarConfiguration,
			int nRows, int nColumns, int nNumberOfFrames, int nBitsAllocated, int nBitsStored, int nHighBit, int nPixelRepresentation,
			float fWindowCenter, float fWindowWidth, float fRescaleIntercept, float fRescaleSlope);
#endif //IMGPROC_NO_DCMTK

		/**
		 * @return the dataset if the DICOM object being processed
		 */
		inline DcmDataset * getDataset() {
			return m_dataset;
		}

		/**
		 * Retrieves the columns in the image(s) stored in the DICOM object
		 *
		 * @return the columns in the image(s) stored in the DICOM objec
		 */
		inline Uint16 getColumns() const {
			return m_columns;
		}

		/**
		 * Retrieves the rows in the image(s) stored in the DICOM object
		 *
		 * @return the rows in the image(s) stored in the DICOM objec
		 */
		inline Uint16 getRows() const {
			return m_rows;
		}

		/**
		 * Retrieves Bits Allocated
		 *
		 * @return Bits Allocated
		 */
		inline Uint16 getBitsAllocated() const {
			return m_bitsAllocated;
		}

		/**
		 * Retrieves Bits Stored
		 *
		 * @return Bits Stored
		 */
		inline Uint16 getBitsStored() const {
			return m_bitsStored;
		}

		/**
		 * Retrieves High Bit
		 *
		 * @return High Bit
		 */
		inline Uint16 getHighBit() const {
			return m_highBit;
		}

		/**
		 * Retrieves Pixel Representation
		 *
		 * @return Pixel Representation
		 */
		inline Uint16 getPixelRepresentation() const {
			return m_pixelRepresentation;
		}

		/**
		 * Retrieves Planar Configuration
		 *
		 * @return Planar Configuration
		 */
		inline Uint16 getPlanarConfiguration() const {
			return m_planarConfiguration;
		}

		/**
		 * Retrieves Samples Per Pixel
		 *
		 * @return Samples Per Pixel
		 */
		inline Uint16 getSamplesPerPixel() const {
			return m_samplesPerPixel;
		}

		/**
		 * Retrieves Number of Frames
		 *
		 * @return Number of Frames
		 */
		inline Uint16 getFrameNumber() const {
			return m_frameNumber;
		}

		/**
		 * Retrieves Rescale Intercept
		 *
		 * @return Rescale Intercept
		 */
		inline double getRescaleIntercept() const {
			return m_rescaleIntercept;
		}

		/**
		 * Retrieves Rescale Slope
		 *
		 * @return Rescale Slope
		 */
		inline double getRescaleSlope() const {
			return m_rescaleSlope;
		}

		/**
		 * Retrieves Window Center
		 *
		 * @return Window Center
		 */
		inline double getWindowCenter() const {
			return m_windowCenter;
		}

		/**
		 * Retrieves Window Width
		 *
		 * @return Window Width
		 */
		inline double getWindowWidth() const {
			return m_windowWidth;
		}

		/**
		 * Retrieves Pixel Spacing X
		 *
		 * @return Pixel Spacing X
		 */
		inline double getPixelSpacingX() const {
			return m_pixelSpacingX;
		}

		/**
		 * Retrieves Pixel Spacing Y
		 *
		 * @return Pixel Spacing Y
		 */
		inline double getPixelSpacingY() const {
			return m_pixelSpacingY;
		}

		/**
		 * Retrieves Photometric Interpretation
		 *
		 * @return Photometric Interpretation
		 */
		inline PI_TYPE getPhotometricInterpretation() const {
			return m_photometricInterpretation;
		}

		/**
		 * Retrieves Image Type
		 *
		 * @return Image Type
		 */
		inline ImageType getImageType() const {
			return m_imageType;
		}

		/**
		 * Retrieves the histogram analyzer
		 *
		 * @return  the histogram analyzer
		 */
		inline CHistogram * getHistogram() {
			return &m_histogram;
		}

		/**
		 * Retrieves Modality
		 *
		 * @return Modality
		 */
		inline const char * getModality() const {
			return m_modality.c_str();
		}

		/**
		 * Indicates whether the raw image data should be inverted
		 * (SuperDASM 1.0 workaround)
		 *
		 * @return OFTrue if and only if the raw image data should be inverted
		 */
		inline OFBool isInverse() const {
			return m_isInverse;
		}

		/**
		 * Retrieves Image Memory manager
		 *
		 * @return Image Memory manager
		 */
		inline CImageMemory * getImageMemory() {
			return m_imageMemory;
		}

		/**
		 * Sets Pixel Representation
		 *
		 * @param pixelRepresentation  New Pixel Representation
		 */
		inline void setPixelRepresentation( Uint16 pixelRepresentation ) {
			m_pixelRepresentation = pixelRepresentation;
		}

		/**
		 * Adds actual bits used value to Actual Bits Used array
		 *
		 * @param actualBitsUsed   value to be added to the Actual Bits Used array
		 */
		void setActualBitsUsed( int actualBitsUsed );

		/**
		 * Adds PB minimum value to PB Min Value array
		 *
		 * @param minValue   value to be added to the PB Min Value array
		 */
		void setPBMinValue( Sint32 minValue );

		/**
		 * Adds PB maximum value to PB Max Value array
		 *
		 * @param maxValue   value to be added to the PB Max Value array
		 */
		void setPBMaxValue( Sint32 maxValue );

		/**
		 * Adds window value to Best Window array
		 *
		 * @param window   window value to be added to the Best Window array
		 * @param window2  window value counted aggressive
		 */
		void setPBPrefWindowWidth( Sint32 window, Sint32 window2 );

		/**
		 * Adds level value to Best Level array
		 *
		 * @param level   level value to be added to the Best Level array
		 * @param level2  level value counted aggressive
		 */
		void setPBPrefWindowCenter( Sint32 level, Sint32 level2 );

		/**
		 * Adds reference values, Best WL values, PB Min/Max values to the dataset
		 */
		void addReferenceValues();

#ifndef	IMGPROC_NO_DCMTK
	protected:

		static bool fTagParserPreprocess( IDicomTagParser2 * obj );


		static void fTag2DCM_PixelData( IDicomTagParser2 * obj );
		static void fTag2DCM_EncapsulatedDocument( IDicomTagParser2 * obj );
		static void fTag2DCM_IconImageSequence( IDicomTagParser2 * obj );
		static void fTag2DCM_Modality( IDicomTagParser2 * obj );
		static void fTag2DCM_SecondaryCaptureDeviceManufacturerModelName( IDicomTagParser2 * obj );
		static void fTag2DCM_SecondaryCaptureDeviceSoftwareVersions( IDicomTagParser2 * obj );
		static void fTag2DCM_SamplesPerPixel( IDicomTagParser2 * obj );
		static void fTag2DCM_PhotometricInterpretation( IDicomTagParser2 * obj );
		static void fTag2DCM_Rows( IDicomTagParser2 * obj );
		static void fTag2DCM_Columns( IDicomTagParser2 * obj );
		static void fTag2DCM_BitsAllocated( IDicomTagParser2 * obj );
		static void fTag2DCM_BitsStored( IDicomTagParser2 * obj );
		static void fTag2DCM_HighBit( IDicomTagParser2 * obj );
		static void fTag2DCM_PixelRepresentation( IDicomTagParser2 * obj );
		static void fTag2DCM_PlanarConfiguration( IDicomTagParser2 * obj );
		static void fTag2DCM_NumberOfFrames( IDicomTagParser2 * obj );
		static void fTag2DCM_NumberOfSlices( IDicomTagParser2 * obj );
		static void fTag2DCM_PixelSpacing( IDicomTagParser2 * obj );
		static void fTag2DCM_WindowCenter( IDicomTagParser2 * obj );
		static void fTag2DCM_WindowWidth( IDicomTagParser2 * obj );
		static void fTag2DCM_RescaleIntercept( IDicomTagParser2 * obj );
		static void fTag2DCM_RescaleSlope( IDicomTagParser2 * obj );
		static void fTag2DCM_BluePaletteColorLookupTableData( IDicomTagParser2 * obj );
		static void fTag2DCM_GreenPaletteColorLookupTableData( IDicomTagParser2 * obj );
		static void fTag2DCM_RedPaletteColorLookupTableData( IDicomTagParser2 * obj );
		static void fTag2DCM_BluePaletteColorLookupTableDescriptor( IDicomTagParser2 * obj );
		static void fTag2DCM_GreenPaletteColorLookupTableDescriptor( IDicomTagParser2 * obj );
		static void fTag2DCM_RedPaletteColorLookupTableDescriptor( IDicomTagParser2 * obj );
		static void fTagPIXEL_DATA_REFERENCE( IDicomTagParser2 * obj );
		static void fTag2DCM_OverlayRows( IDicomTagParser2 * obj );
		static void fTag2DCM_OverlayColumns( IDicomTagParser2 * obj );
		static void fTag2DCM_OverlayType( IDicomTagParser2 * obj );
		static void fTag2DCM_OverlayOrigin( IDicomTagParser2 * obj );
		static void fTag2DCM_OverlayBitsAllocated( IDicomTagParser2 * obj );
		static void fTag2DCM_OverlayBitPosition( IDicomTagParser2 * obj );
		static void fTag2DCM_OverlayData( IDicomTagParser2 * obj );

	// Stuff to report back the progress of the process. Needed for herelod, so
	// it won't timeout because it knows that something is happening during the
	// time of processing.
	protected:
		IImageProcessProgress* m_pProgressCallback;

	public:
		void setProgressCallback(IImageProcessProgress* pProgressCallback);
#endif //IMGPROC_NO_DCMTK

};

#endif // __IMAGE_PROCESS_H_INCLUDED__
