/*
 * Component:  common
 *
 * Purpose:
 *	    Header file for overlay processing
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */
#ifndef __OVERLAY_H_INCLUDED__
#define __OVERLAY_H_INCLUDED__

#ifndef IMGPROC_NO_DCMTK
#include "dcmtk/ofstd/ofstd.h"
#include "dcmtk/dcmdata/dctk.h"
#else
#include "imgproc/nodcmtk.h"
#endif //IMGPROC_NO_DCMTK

/** Overlay Type */
typedef enum {
	GRAPHICS_OVERLAY,
	ROI_OVERLAY
} OVERLAY_TYPE;

/** Overlay Read Mode */
typedef enum {
	OVERLAY_UNINITIALIZED,
	OVERLAY_6000,
	OVERLAY_PIXEL
} OVERLAY_READ_MODE;


/**
* Class holds information on an overlay stored in a DICOM object.
*/
class COverlay {
protected:

	/** Reference identifier (reference left in the 'herelt' DICOM file) */
	OFString m_id;

	/** Bytestream holding overlay pixels */
	Uint8 * m_data;

	/** Length of m_data */
	Uint32 m_length;

	/** number of rows (height) of the overlay */
	Uint16 m_rows;

	/** number of columns (with) of the overlay */
	Uint16 m_columns;

	/** Row of image where the topleft overlay pixel should be rendered (row numbers are starting with 1) */
	Sint16 m_originRow;

	/** Column of image where the topleft overlay pixel should be rendered (column numbers are starting with 1) */
	Sint16 m_originColumn;

	/** Bits Allocated for Overlay data */
	Uint16 m_bitsAllocated;

	/** Bit number where the overlay is stored in pixel data */
	Uint16 m_overlayBit;

	/** Overlay type (graphics/ROI) */
	OVERLAY_TYPE m_type;

	/** Mode for filling overlay from DICOM file */
	OVERLAY_READ_MODE m_mode;

	/** Number of pixels in the overlay */
	Uint32 m_numberOfPixels;

	/** Number of current pixel bit in current byte (helper for OVERLAY_PIXEL mode) */
	Uint32 m_pixelBitCounter;

	/** Number of current byte (helper for OVERLAY_PIXEL and OVERLAY_6000 modes) */
	Uint32 m_byteCounter;

	/** Pointer to current byte (helper for OVERLAY_PIXEL and OVERLAY_6000 modes) */
	Uint8 * m_currentByte;

	/** Error code of last error */
	int m_error;

public:

	/**
	* Constructor for creating a new overlay object
	*
	* @param id      Reference identifier (reference left in the 'herelt' DICOM file)
	*/
	COverlay(const OFString & id);

	/**
	* Destructor for destroying an overlay object
	*/
	~COverlay();


	/**
	* Retrieves the reference identifier (reference left in the 'herelt' DICOM file)
	*
	* @return the reference identifier (reference left in the 'herelt' DICOM file)
	*/
	inline const OFString & getId() { return m_id; }

	/**
	* Retrieves the overlay pixel data
	*
	* @return the overlay pixel data
	*/
	inline const Uint8 * getData() { return m_data; }

	/**
	* Retrieves the length overlay pixel data
	*
	* @return the length overlay pixel data
	*/
	inline Uint32 getLength() { return m_length; }

	/**
	* Sets the number of rows in the overlay
	*
	* @param rows       Number of rows in the overlay
	*/
	inline void setRows(Uint16 rows) { m_rows = rows; }

	/**
	* Retrieves the number of rows in the overlay
	*
	* @return the number of rows in the overlay
	*/
	inline Uint16 getRows() { return m_rows; }

	/**
	* Sets the number of columns in the overlay
	*
	* @param columns    Number of columns in the overlay
	*/
	inline void setColumns(Uint16 columns) { m_columns = columns; }

	/**
	* Retrieves the number of columns in the overlay
	*
	* @return the number of columns in the overlay
	*/
	inline Uint16 getColumns() { return m_columns; }

	/**
	* Sets the number of bits allocated for the overlay
	*
	* @param bitsAllocated    The number of bits allocated for the overlay
	*/
	inline void setBitsAllocated(Uint16 bitsAllocated) { m_bitsAllocated = bitsAllocated; }

	/**
	* Retrieves the number of bits allocated for the overlay
	*
	* @return the number of bits allocated for the overlay
	*/
	inline Uint16 setBitsAllocated() { return m_bitsAllocated; }

	/**
	* Sets the serial number of bit where the overlay is stored in pixel data
	*
	* @param overlayBit    The serial number of bit where the overlay is stored in pixel data
	*/
	inline void setOverlayBit(Uint16 overlayBit) { m_overlayBit = overlayBit; }

	/**
	* Retrieves the serial number of bit where the overlay is stored in pixel data
	*
	* @return the serial number of bit where the overlay is stored in pixel data
	*/
	inline Uint16 getOverlayBit() { return m_overlayBit; }

	/**
	* Sets the row of image where the topleft overlay pixel should be rendered
	* (row numbers are starting with 1)
	*
	* @param originRow       Row number in image where the top of the overlay should be drawed
	*/
	inline void setOriginRow(Sint16 originRow) { m_originRow = originRow; }

	/**
	* Retrieves the row of image where the topleft overlay pixel should be rendered
	* (row numbers are starting with 1)
	*
	* @return row number in image where the top of the overlay should be drawed
	*/
	inline Sint16 getOriginRow() { return m_originRow; }

	/**
	* Sets the column of image where the topleft overlay pixel should be rendered
	* (column numbers are starting with 1)
	*
	* @param originColumn    Column number in image where the left of the overlay should be drawed
	*/
	inline void setOriginColumn(Sint16 originColumn) { m_originColumn = originColumn; }

	/**
	* Retrieves the column of image where the topleft overlay pixel should be rendered
	* (column numbers are starting with 1)
	*
	* @return column number in image where the left of the overlay should be drawed
	*/
	inline Sint16 getOriginColumn() { return m_originColumn; }

	/**
	* Sets overlay type (graphics/ROI)
	*
	* @param type    Overlay type
	*/
	inline void setOverlayType(OVERLAY_TYPE type) { m_type = type; }

	/**
	* Retrieves overlay type (graphics/ROI)
	*
	* @return overlay type
	*/
	inline OVERLAY_TYPE getOverlayType() { return m_type; }


	/**
	* Initialize for filling overlay data.
	* Will allocate rows x columns / 8 + 1 bytes for the overlay data.
	* Counters and helper variables for filling the overlay data will be
	* initialized here.
	*
	* @param length  Length of the overlay data
	* @param mode    Mode of filling:
	*                  OVERLAY_UNINITIALIZED : mode will be autodetected from Bits Allocated and Overlay Bit values (default)
	*                  OVERLAY_6000          : Overlay data will be filled from 60xx tags using putData()
	*                  OVERLAY_PIXEL         : Overlay data will be filled from pixelData using setNextPixel()
	*
	* @return 0 - OK
	*         1 - Invalid value for mode
	*         2 - Already contains overlay data or initData() called twice
	*         3 - length is 0
	*         4 - Not enough memory
	*/
	int initData(Uint32 length, OVERLAY_READ_MODE mode = OVERLAY_UNINITIALIZED);

	/**
	* Put pixels to overlay data. Can be used in cases the overlay data is received
	* in 60xx,0300.
	*
	* Important:
	*   putData() and setNextPixel should not be used together
	*
	* @param data      pixels to be put to overlay data
	* @param length    length of pixel data (in bytes)
	*
	* @return 0 - OK
	*         1 - There was an error earlier or invalid mode (see initData)
	*         5 - NULL value
	*         6 - data received in argument too long
	*/
	int putData(const Uint8 * data, Uint32 length);

	/**
	* Sets the next pixel in overlay data
	*
	* Important:
	*   putData() and setNextPixel should not be used together
	*
	* @param pixel   pixel value of the next pixel in the overlay
	*
	* @return 0 - OK
	*         1 - There was an error earlier or invalid mode (see initData)
	*         6 - data received in argument too long
	*/
	inline int setNextPixel(bool pixel) {
		if (m_error != 0 || m_mode != OVERLAY_PIXEL) {
			return 1;
		}

		if (m_byteCounter >= m_length) {
			return m_error = 6;
		}

		*m_currentByte |= ((int) pixel << m_pixelBitCounter++);
		if (m_pixelBitCounter == 8) {
			m_pixelBitCounter = 0;
			m_byteCounter++;
			m_currentByte++;
		}
		return 0;
	}

	/**
	* Signals that filling overlay data is finished. Will check if the
	* overlaydata is valid (no more or less pixels than required)
	*
	* @return 0 - OK
	*         1 - invalid mode (uninitialized)
	*         6 - there were more pixel then rows x columns
	*         7 - pixels stored in overlay data is less then rows x columns
	*/
	int finalizeData();

	/**
	* Writes the overlay into a file
	*
	* @param fileName   Name of the overlay file
	*
	* @return 0 on success, >0 on error
	*/
	int writeToFile(const char * fileName);

};

#define MAX_NUMBER_OF_OVERLAYS 16

/**
* Array of overlays stored in DICOM object
*/
class COverlayArray {

protected:

	/** Array of overlays read from a DICOM object. A DICOM object may contain at most 16 overlays per documentation */
	COverlay * m_overlay[MAX_NUMBER_OF_OVERLAYS];

	/** Study Instance UID of study the DICOM object belongs to */
	OFString m_study;

	/** File name of the DICOM object */
	OFString m_file;

public:

	/**
	* Constructor for creating a new cOverlayArray object
	*/
	COverlayArray();

	/**
	* Destructor for destroying the COverlayArray object
	*/
	~COverlayArray();

	/**
	* Creates a new overlay object and add it to the array.
	*
	* @return the index of the newly created overlay object or -1 on error
	*/
	int createNewOverlay();

	inline int getNumberOfMaxOverlays() const {
		return MAX_NUMBER_OF_OVERLAYS;
	}

	/**
	* Retrieves the specified overlay object
	*
	* @param ndx   Index of the Overlay object to be retrieved
	*
	* @return the overlay object or NULL if does not exist
	*/
	inline COverlay * getOverlay(int ndx) {
		return m_overlay[ndx];
	}

	/**
	* Sets the number of rows in the overlay (the value of (60xx,0010) tag or image rows)
	* for a specified overlay.
	*
	* @paran ndx      Index of the overlay
	* @param rows     Number of rows in the overlay
	*
	* @return OFTrue on success or OFFalse if there is no overlay at ndx
	*
	* @see COverlay#setRows(Uint16 rows)
	*/
	inline OFBool setOverlayRows(int ndx, Uint16 rows) {
		if (m_overlay[ndx] == NULL)
			return OFFalse;
		m_overlay[ndx]->setRows(rows);
		return OFTrue;
	}

	/**
	* Sets the number of columns in the overlay (the value of (60xx,0011) tag or image columns)
	* for a specified overlay.
	*
	* @paran ndx      Index of the overlay
	* @param columns  Number of columns in the overlay
	*
	* @return OFTrue on success or OFFalse if there is no overlay at ndx
	*
	* @see COverlay#setColumns(Uint16 columns)
	*/
	inline OFBool setOverlayColumns(int ndx, Uint16 columns) {
		if (m_overlay[ndx] == NULL)
			return OFFalse;
		m_overlay[ndx]->setColumns(columns);
		return OFTrue;
	}

	/**
	* Sets the number of bits allocated for the specified overlay
	*
	* @paran ndx              Index of the overlay
	* @param bitsAllocated    The number of bits allocated for the overlay
	*
	* @return OFTrue on success or OFFalse if there is no overlay at ndx
	*
	* @see COverlay#setBitsAllocated(Uint16 columns)
	*/
	inline OFBool setOverlayBitsAllocated(int ndx, Uint16 bitsAllocated) {
		if (m_overlay[ndx] == NULL)
			return OFFalse;
		m_overlay[ndx]->setBitsAllocated(bitsAllocated);
		return OFTrue;
	}

	/**
	* Sets the serial number of bit where the overlay is stored in pixel data for a specified overlay
	*
	* @paran ndx           Index of the overlay
	* @param overlayBit    The serial number of bit where the overlay is stored in pixel data
	*
	* @return OFTrue on success or OFFalse if there is no overlay at ndx
	*
	* @see COverlay#setOverlayBit(Uint16 columns)
	*/
	inline OFBool setOverlayBit(int ndx, Uint16 overlayBit) {
		if (m_overlay[ndx] == NULL)
			return OFFalse;
		m_overlay[ndx]->setOverlayBit(overlayBit);
		return OFTrue;
	}

	/**
	* Sets the row of image where the topleft overlay pixel should be rendered
	* (column numbers are starting with 1) for a specified overlay.
	*
	* @paran ndx      Index of the overlay
	* @param columns  Row number in image where the top of the overlay should be drawed
	*
	* @return OFTrue on success or OFFalse if there is no overlay at ndx
	*
	* @see COverlay#setOriginRow(Sint16 originRow)
	*/
	inline OFBool setOverlayOriginRow(int ndx, Sint16 originRow) {
		if (m_overlay[ndx] == NULL)
			return OFFalse;
		m_overlay[ndx]->setOriginRow(originRow);
		return OFTrue;
	}

	/**
	* Sets the column of image where the topleft overlay pixel should be rendered
	* (column numbers are starting with 1) for a specified overlay.
	*
	* @paran ndx      Index of the overlay
	* @param columns  Column number in image where the top of the overlay should be drawed
	*
	* @return OFTrue on success or OFFalse if there is no overlay at ndx
	*
	* @see COverlay#setOriginColumn(Sint16 originColumn)
	*/
	inline OFBool setOverlayOriginColumn(int ndx, Sint16 originColumn) {
		if (m_overlay[ndx] == NULL)
			return OFFalse;
		m_overlay[ndx]->setOriginColumn(originColumn);
		return OFTrue;
	}

	/**
	* Sets the overlay type for a specified overlay.
	*
	* @paran ndx      Index of the overlay
	* @param type     Overlay type
	*
	* @return OFTrue on success or OFFalse if there is no overlay at ndx
	*
	* @see COverlay#setOverlayType(Sint16 originColumn)
	*/
	inline OFBool setOverlayType(int ndx, OVERLAY_TYPE type) {
		if (m_overlay[ndx] == NULL)
			return OFFalse;
		m_overlay[ndx]->setOverlayType(type);
		return OFTrue;
	}

	/**
	* Initialize the specified overlay object to put overlay data from
	* (60xx,0300) DICOM tag.
	*
	* @param ndx       Index of the overlay
	* @param length    The length of the overlay data (the (60xx,0300) DICOM tag)
	*
	* @return 0 on success, -1 if there is no such overlay or error code
	*
	* @see initPixelData(Uint32 length, Uint16 & mask)
	* @see putOverlayData(const Uint8 * data, Uint32 length)
	* @see COverlay#initData(Uint32 length, OVERLAY_READ_MODE mode)
	*/
 	inline int initOverlayData(int ndx, Uint32 length) {
 		if (m_overlay[ndx] == NULL)
 			return -1;
 		return m_overlay[ndx]->initData(length, OVERLAY_6000);
 	}

	/**
	* Initialize overlays to put overlay data from pixel data.
	* All overlays waiting for pixel data will be initialized.
	*
	* A mask will be filled which cen be used to collect overlay info from pixel data.
	*
	* @param rows      Number of rows in pixel data
	* @param columns   Number of columns in pixel data
	* @param mask      Output parameter for bit mask to be used for get overlay data from pixel data
	*
	* @return 0 on success or >0 in case of any error
	*
	* @see initOverlayData(int ndx, Uint32 length)
	* @see COverlay#initData(Uint32 length, OVERLAY_READ_MODE mode)
	*/
 	inline int initPixelData(Uint16 rows, Uint16 columns, Uint16 & mask) {
 		int error = 0;
		Uint32 pixelNum = rows * columns;
		Uint32 length = pixelNum / 8;

		if (length % 8 != 0) {
			length++;
		}

 		for (int i = 0 ; i < 16; i++) {
 			if (m_overlay[i] != NULL && m_overlay[i]->getData() == NULL) {
				m_overlay[i]->setRows(rows);
				m_overlay[i]->setColumns(columns);
 				error += m_overlay[i]->initData(length);
				mask |= (1 << m_overlay[i]->getOverlayBit());
			}
 		}
		return error;
 	}

	/**
	* Puts overlay data from (60xx,0300) DICOM tag to the specified overlay object.
	* It may happen that entire overlay data will be read in more parts.
	*
	* @param ndx       Index of the overlay
	* @param data      Overlay data part
	* @param length    The length of the overlay data part
	*
	* @return 0 on success or -1 if there is no such overlay or error code
	*
	* @see initOverlayData(int ndx, Uint32 length)
	* @see COverlay#putData(const Uint8 * data, Uint32 length)
	*/

	inline int putOverlayData(int ndx, const Uint8 * data, Uint32 length) {
		if (m_overlay[ndx] == NULL)
			return -1;
		return m_overlay[ndx]->putData(data, length);
	}

	/**
	* Sets the next pixel in each overlay.
	*
	* Iterates through the overlays and sets the next pixel value in them with the
	* overlayBit-th of pixels
	*
	* @param pixels   contains overlay pixel values for each overlay
	*
	* @return 0  - OK
	*         >0 - If an error occures
	*
	* @see initPixelData(Uint32 length, Uint16 & mask)
	* @see COverlay#setNextPixel(bool)
	*/
	inline int setNextPixels(Uint16 pixels) {
		int error = 0;
		for (int i = 0 ; i < 16; i++) {
			if (m_overlay[i] != NULL && (m_overlay[i]->getOverlayBit() != 0)) {
				error += m_overlay[i]->setNextPixel( (pixels & (1 << m_overlay[i]->getOverlayBit())) != 0 );
			}
		}
		return error;
	}

	/**
	* Signals that filling overlay data is finished. Will iterate through all overlays
	* and finalize them.
	*
	* @return 0  - OK
	*         >0 - If an error occures
	*
	* @see COverlay#finalizeData()
	*/
	inline int finalizeOverlays() {
		int error = 0;
		for (int i = 0 ; i < 16; i++) {
			if (m_overlay[i] != NULL)
				error += m_overlay[i]->finalizeData();
		}
		return error;
	}

	/**
	* Write overlays into files. The overlay index and the ".ovl" extension will be
	* appended to the prefix to create the overlay file names.
	*
	* @param prefix   Overlay file name prefix
	*
	* @return 0 on success, >0 on error
	*/
	int writeOverlays(const char * prefix);
};

#endif // __OVERLAY_H_INCLUDED__
