#ifndef IMUTILS_INCLUDE_IMUTILS_BENCHTIMER_H_
#define IMUTILS_INCLUDE_IMUTILS_BENCHTIMER_H_

#include <vector>
#include <string>
#include <chrono>
#include <iostream>

#include "imutils/IMLogger.h"

namespace erad::imutils {

/**
 * Utility class for measuring execution times for a part (or parts) of a code.
 * It can calculate and collect elapsed times between benches, and also measures overall running time.
 * The collected values will be printed or logged in the destructor.
 * The overall time is the elapsed time between the construction and the destruction of the benchmark object
 *
 * Typical usage:
 *      {
 *         BenchTimer bt( ... );
 *
 *         ... do something
 *
 *         bt.benchTime()
 *
 *         ... do something
 *
 *         bt.benchTime()
 *      }
 */
class BenchTimer {
		/// the name (label) of the bench
		std::vector<std::string> names;
		/// elapsed time for a bench (since the previous bench)
		std::vector<long> times;
		/// the start time
		std::chrono::time_point<std::chrono::high_resolution_clock> start;
		/// timestamp of the latest bench
		std::chrono::time_point<std::chrono::high_resolution_clock> latest;

		/// optional pointer to a `logger` to add the collected values to a log entry
		IMLogger* logger;
		/// optional pointer to an output stream to write the collected values to the stream
		std::ostream* output;
		/// message to be prepended to the log entry or the output
		std::string message;

	public:

		/**
		 * Creates a new `BenchTimer` object.
		 */
		BenchTimer();

		/**
		 * Creates a new `BenchTimer` object to log the results.
		 * An `info` level log entry will be automatically logged at the end of the measurement
		 * (by the destructor)
		 *
		 * @param logger the logger object
		 * @param message the message to be logged along with the collected values.
		 */
		BenchTimer( IMLogger* logger, const std::string& message );

		/**
		 * Creates a new `BenchTimer` object to print the results to an output stream.
		 * The results will be automatically printed to the output stream at the end of the measurement
		 * (by the destructor)
		 *
		 * @param output the output stream
		 * @param message the message to be printed along with the collected values.
		 */
		BenchTimer( std::ostream* output, const std::string& message );
		~BenchTimer();

		/* Copy and move constructors are disabled */

		BenchTimer( const BenchTimer& orig ) = delete;
		BenchTimer( BenchTimer&& orig ) = delete;

		/* Copy and move assignment operators are also disabled */

		BenchTimer& operator=( const BenchTimer& ) = delete;
		BenchTimer& operator=( BenchTimer&& ) = delete;

		/**
		 * Sets a logger to log the results.
		 * An `info` level log entry will be automatically logged at the end of the measurement
		 * (by the destructor)
		 *
		 * @param logger the logger object
		 * @param message the message to be logged along with the collected values.
		 */
		inline void setLogger( IMLogger* logger, const std::string& message ) {
			this->logger = logger;
			this->message = message;
		}

		/**
		 * Sets an output stream to print the results.
		 * The results will be automatically printed to the output stream at the end of the measurement
		 * (by the destructor)
		 *
		 * @param output the output stream
		 * @param message the message to be printed along with the collected values.
		 */
		inline void setOutput( std::ostream* output, const std::string& message ) {
			this->output = output;
			this->message = message;
		}

		/**
		 * Stores a new bench. Calculates the elapsed time since to latest bench and and stores the result,
		 *
		 * @param name the name of the bench
		 */
		inline void benchTime( const std::string& name ) {
			std::chrono::time_point<std::chrono::high_resolution_clock> now = std::chrono::high_resolution_clock::now();
			names.emplace_back( name );
			times.emplace_back( std::chrono::duration_cast<std::chrono::nanoseconds>( now - latest ).count() );
			latest = now;
		}

		/**
		 * Calculates the overall time and creates a string representation of the results.
		 * @return the results as a `string`.
		 */
		std::string toString();

		/**
		 * Prints the results to an output stream
		 * @param os the output stream
		 * @param bt the `BenchTimer` object to be printed
		 * @return the output stream (`os`)
		 */
		friend std::ostream& operator<<( std::ostream& os, const BenchTimer& bt );
};

} // namespace

#endif /* IMUTILS_INCLUDE_IMUTILS_BENCHTIMER_H_ */
