/*
 * Component:  utils
 *
 * Purpose:
 *   Base class for exception handling
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 *
 */

#ifndef __IM_EXCEPTION_H_INCLUDED__
#define __IM_EXCEPTION_H_INCLUDED__

#include <stdio.h>

#include <string>
#include <ostream>
using namespace std;

/**
 * Exception class can be used in PB programs
 */
class IMException {
	protected:

		/** Error message of the exception */
		string m_message;

		/** Error code */
		int m_errorCode;

		/** Initialize */
		void init();

	public:

		/**
		 * Default constructor
		 */
		IMException();

		/**
		 * Create new exception with a specified error code
		 * (error message will be empty)
		 *
		 * @param errorCode   error code for the exception
		 */
		IMException( int errorCode );

		/**
		 * Create new exception with a specified message
		 * (error code will be 1)
		 *
		 * @param mask   printf like mask
		 * @param ...    positional parameters for the mask
		 */
		IMException( const char * mask, ... );

		/**
		 * Create new exception with specified message and error code
		 */
		IMException( int errorCode, const char * mask, ... );

		/**
		 * Create new exception with specified message and error code
		 */
		IMException( int errorCode, const string & message );

		/**
		 * Destructor to destroy exception
		 */
		virtual ~IMException();

		/**
		 * Retrieves message of the exception
		 *
		 * @return message of the exception
		 */
		inline const char* getMessage() const {
			return m_message.c_str();
		}

		/**
		 * Retrieves message of the exception
		 *
		 * @return message of the exception
		 */
		inline const string & getOFMessage() {
			return m_message;
		}

		/**
		 * Retrieves error code of the exception
		 *
		 * @return error code of the exception
		 */
		inline int getErrorCode() {
			return m_errorCode;
		}

		/**
		 * Print the exception into a file descriptor
		 *
		 * @param f  file descriptor to print into
		 */
		void print( FILE * f = stdout );

		/**
		 * Print the exception into an output stream
		 *
		 * @param out  output stream to print into
		 */
		void print( ostream out );

		/**
		 * Converts the exception into a string representation
		 * The string returned by toString() should bee freed up.
		 *
		 * @return the string representation of the exception
		 */
		char * toString();

		/**
		 * Converts the exception into a string representation.
		 *
		 * @param str  output parameter for the string representation of the exception
		 */
		virtual void toString( string & str );
};

/**
 * Exception class can be used when a timeout is reached
 */
class IMTimeoutException : public IMException {
	public:
		/**
		 * Create new timeout exception with specified message and error code
		 */
		IMTimeoutException( int errorCode, const string & message );

		/**
		 * Converts the timeout exception into a string representation.
		 *
		 * @param str  output parameter for the string representation of the timeout exception
		 */
		virtual void toString( string & str );
};

/**
 * Exception class can be used when Input/Output errors occured
 */
class IMIOException : public IMException {
	public:
		/**
		 * Create new io exception with specified message and error code
		 */
		IMIOException( int errorCode, const string & message );

		/**
		 * Converts the io exception into a string representation.
		 *
		 * @param str  output parameter for the string representation of the io exception
		 */
		virtual void toString( string & str );
};

class IMAssertionException : public IMException {
	private:
		string m_assert;
		string m_file;
		long m_line;
		bool m_hasMessage;

	public:

		/**
		 * Create new assertion exception with assert information and specified message
		 *
		 * @param assert    Assert condition in text format
		 * @param line      Line number in the source file where the assertion failed
		 * @param file      Name of the source file
		 * @param message   printf-like string mask with var args
		 */
		IMAssertionException( const char * assert, long line, const char * file, const char * mask, ... );

		/**
		 * Create new assertion exception with assert information and specified message
		 *
		 * @param assert    Assert condition in text format
		 * @param line      Line number in the source file where the assertion failed
		 * @param file      Name of the source file
		 */
		IMAssertionException( const char * assert, long line, const char * file );

		/**
		 * Converts the assert exception into a string representation.
		 *
		 * @param str  output parameter for the string representation of the assert exception
		 */
		virtual void toString( string & str );
};

#define IM_ASSERT(x)   if (!(x)) { throw IMAssertionException(#x, __LINE__, __FILE__); }

#define IM_ASSERT_MSG(x, ...) if (!(x)) { throw IMAssertionException(#x, __LINE__, __FILE__, __VA_ARGS__); }

#endif // __IM_EXCEPTION_H_INCLUDED__
