/*
 * Component:  utils
 *
 * Purpose: 
 *   Header file for semaphore implementation to get exclusive lock on a resource
 *
 */

#ifndef SEMAPHORE_H
#define SEMAPHORE_H

#include <string>
using namespace std;


/**
 * Creates/removes exclusive lock on a resource.
 */
class IMSemaphore {
	protected:

		/** resource identifier */
		string m_resourceId;

		/** directory name where the lockfile should be created */
		string m_lockDir;

		/** indicates whether to release the lock when the semaphore object destroyed (true by default)  */
		bool m_releaseOnDestroy;

		/** indicates whether this semaphore locks the resource or not */
		bool m_isLocked;

		/** indicates whether unlock is forced even if the lock was created by another application */
		bool m_isUnlockForced;

		/**
		 * Attempts to lock the resource
		 *
		 * @return true if and only if the resource was successfully locked
		 *
		 * @throws IMIOException if semaphore file did not exist and could not be created.
		 */
		virtual bool lockResource();

	public:

		/**
		 * Constructor for creating a new IMSemaphore object on a resource
		 *
		 * @param resourceId       identifies the resource to be semaphored
		 */
		IMSemaphore( const string & resourceId );

		/**
		 * Destructor to destroy IMSemaphor object
		 */
		virtual ~IMSemaphore();

		/**
		 * Tries to lock the resource. If the resource could not be locked, then lock will be retried
		 * several times until the timeout is expired.
		 *
		 * @param timeout   timeout in seconds (default is 30)
		 *
		 * @return true if and only if the resource was successfully locked
		 *
		 * @throws IMTimeoutException - if the timeout is reached
		 *         IMIOException if semaphore file did not exist and could not be created.
		 */
		bool lock( int timeout = 30 );


		/**
		 * Tries to lock the resource. If the resource could not be locked, then lock will be retried
		 * several times until the timeout is expired.
		 *
		 * @param timeout_ms   timeout in milliseconds
		 *
		 * @param tick_ms		initial retry tick size in milliseconds (default is 10.0)
		 *
		 * @param tickScale	tick increment factor (default is 1.5)
		 *
		 * @param tickMax		maximum tick (default is 1000.0 = 1 second)
		 *
		 * @param tickJitter	random tick jitter (default is 0.4 = +/-40%)
		 *
		 * @return true if and only if the resource was successfully locked
		 *
		 * @throws IMTimeoutException - if the timeout is reached
		 *         IMIOException if semaphore file did not exist and could not be created.
		 */
		bool lock_ms( int timeout_ms, float tick_ms = 10.0f, float tickScale = 1.5f, float tickMax = 1000.0f, float tickJitter = 0.4f );

		/**
		 * Removes the lock created by this Semaphore object from the resource
		 *
		 * @throws IMIOException if semaphore file could not be released.
		 */
		virtual void unlock();

		/**
		 * Checks if the resource is locked by this Semaphore object
		 *
		 * @return true if and only if this Semaphore object locks the resource.
		 */
		inline bool isLocked() {
			return m_isLocked;
		}

		/**
		 * Sets a custom lock directory.
		 * The lock directory should be set before any lock get created.
		 *
		 * @param lockDir path to the lock directory
		 */
		inline void setLockDir( const std::string& lockDir ) {
			m_lockDir = lockDir;
		}

		/**
		 * Sets if the lock should be released on destroy
		 *
		 * @param releaseOnDestroy indicates whether a lock should be released when Semaphore is destroyed.
		 */
		inline void setReleaseOnDestroy( bool releaseOnDestroy ) {
			m_releaseOnDestroy = releaseOnDestroy;
		}

		/**
		 * Force unlocking semaphore, even if the lock was created by another application.
		 *
		 * TAKE EXTRA CARE WHEN USING THIS METHOD
		 */
		inline void forceUnlock() {
			m_isUnlockForced = true;
		}

};

#endif
