/*
 * Interface definition to access DICOM object information
 *
 * Also implements the interface for MySQL.
 */

#ifndef IMUTILS_INCLUDE_IMUTILS_OBJECTDB_H_
#define IMUTILS_INCLUDE_IMUTILS_OBJECTDB_H_

#include <string>
#include <list>

#include "imutils/DBConnector.h"

namespace erad::db {

/* ********************************************
 *
 * Helper bundles to pass/return information by the interface methods
 *
 * ********************************************/

/**
 * A bundle to hold identifier attributes of a report object.
 */
class ReportFileInfo {
	public:

		/// filename of the report object
		std::string fileName;
		/// the SOP class UID of the report object
		std::string sopClass;
		/// the SOP instance UID of the report object
		std::string instanceUid;

		/**
		 * Creates a new ReportFileInfo object
		 *
		 * @param fileName filename of the report object
		 * @param sopClass the SOP class UID of the report object
		 * @param instanceUid the SOP instance UID of the report object
		 */
		ReportFileInfo( const std::string& fileName, const std::string& sopClass, const std::string& instanceUid )
				: 	fileName( fileName ),
					sopClass( sopClass ),
					instanceUid( instanceUid ) {
		}

		/* Default, copy and move constructors are disabled */

		ReportFileInfo() = delete;
		ReportFileInfo( const ReportFileInfo& orig ) = delete;
		ReportFileInfo( ReportFileInfo&& orig ) = delete;

		~ReportFileInfo() {
		}

		/* Copy and move assignment operators are also disabled */

		ReportFileInfo& operator=( const ReportFileInfo& ) = delete;
		ReportFileInfo& operator=( ReportFileInfo&& ) = delete;
};


/**
 * A bundle to hold the object UID along with the containing series and file name.
 */
class SeriesObject {
	public:

		/// the series instance UID of the object
		std::string series;

		/// the SOP instance UID of the object
		std::string object;

		/// the file name of the object
		std::string fileName;

		/**
		 * Creates a new SeriesObject object
		 *
		 * @param series the series instance UID of the object
		 * @param object the SOP instance UID of the object
		 * @param fileName the file name of the object
		 */
		SeriesObject( const std::string& series, const std::string& object, const std::string& fileName )
				: 	series( series ),
					object( object ),
					fileName( fileName ) {
		}

		/* Default, copy and move constructors are disabled */

		SeriesObject() = delete;
		SeriesObject( const SeriesObject& orig ) = delete;
		SeriesObject( SeriesObject&& orig ) = default;

		~SeriesObject() {
		}

		/* Copy and move assignment operators are also disabled */

		SeriesObject& operator=( const SeriesObject& ) = delete;
		SeriesObject& operator=( SeriesObject&& ) = delete;
};


/**
 * A bundle to hold identifier attributes of an image object.
 */
class ImageInfo {
	public:

		/// the SOP class UID of the image object
		std::string sopClass;
		/// the SOP instance UID of the image object
		std::string sopInstance;
		/// the series UID of the image object
		std::string series;
		/// number of columns in the image
		int columns = 0;
		/// number of rows in the image
		int rows = 0;
		/// number of image frames
		int frames = 0;


		/**
		 * Creates an empty ImageInfo object
		 */
		ImageInfo() {
		}

		/* Copy and move constructors are disabled */

		ImageInfo( const ImageInfo& orig ) = delete;
		ImageInfo( ImageInfo&& orig ) = default;

		~ImageInfo() {
		}

		/* Copy and move assignment operators are also disabled */

		ImageInfo& operator=( const ImageInfo& ) = delete;
		ImageInfo& operator=( ImageInfo&& ) = delete;
};


/**
 * A bundle to hold object cache information for a study.
 */
class CacheInfo {
	public:

		/// object cache status of the study
		int status = 0;
		/// true indicates the object cache entry was touched recently
		bool isRecent = false;

		/**
		 * Creates an empty CacheInfo object
		 */
		CacheInfo() {
		}

		/* Copy and move constructors are disabled */

		CacheInfo( const CacheInfo& orig ) = delete;
		CacheInfo( CacheInfo&& orig ) = default;

		~CacheInfo() {
		}

		/* Copy and move assignment operators are also disabled */

		CacheInfo& operator=( const CacheInfo& ) = delete;
		CacheInfo& operator=( CacheInfo&& ) = delete;
};


/* ********************************************
 *
 * Interface
 *
 * ********************************************/


/**
 * Interface to get information from the object table
 */
class ObjectDB {

	public:
		virtual ~ObjectDB() {
		}

		/**
		 * Gets the objects from a study.
		 * Lists file names of the objects in the study. All objects will be returned except the PbR.
		 *
		 * Used by the edit interface to apply soft edit changes to objects.
		 *
		 * @param study [in] the study instance UID
		 * @param files [out] output parameter for the object files in the study
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual void getStudyFiles( const std::string& study, std::list<std::string>& files /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets the objects from a series.
		 * Lists file names of the objects in the series. All objects will be returned except the PbR.
		 *
		 * Used by the edit interface to edit series.
		 *
		 * @param study [in] the study instance UID
		 * @param series [in] the series instance UID
		 * @param files [out] output parameter for the object files in the study
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual void getSeriesFiles( const std::string& study, const std::string& series, std::list<std::string>& files /*out*/)
				noexcept ( false ) = 0;

		/**
		 * Gets the file name for an object.
		 *
		 * Used by
		 * - study merge to get the file name of the minsopi object and
		 * - the new edit interface to get the filename for a report object
		 *
		 * @param study [in] the study instance UID of the object
		 * @param sopiuid [in] the SOP instance UID of the object
		 * @param file [out] output parameter for the file name of the object
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the object was found
		 */
		virtual bool getFileByUid( const std::string& study, const std::string& sopiuid, std::string& file /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets the PbR file for a study.
		 *
		 * Used by
		 * - study merge
		 * - report and order manager to load the PbR
		 * - detaching shortcuts
		 *
		 * @param study [in] the study instance UID
		 * @param file [out] output parameter for the file name of the PbR object
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the PbR object was found
		 */
		virtual bool getPbRFile( const std::string& study, std::string& file /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets the PbR series instance uid and sop instance uid for a study.
		 *
		 * Used by
		 * - study merge
		 * - report and order manager to load the PbR
		 * - detaching shortcuts
		 *
		 * @param study [in] the study instance UID
		 * @param series [out] the series instance UID
		 * @param sopInstance [out] the sop instance UID
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the PbR object was found
		 */
		virtual bool getPbRInstance( const std::string& study, std::string& series /*out*/, std::string& sopInstance /*out*/)
			noexcept ( false ) = 0;

		/**
		 * Gets the dictation file(s) for a study.
		 *
		 * Used by the report manager to load the dictation(s).
		 *
		 * @param study [in] the study instance UID
		 * @param file [out] output parameter for the file name of the dictation (PbD) object(s).
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @note currently only file name is used by the caller
		 */
		virtual void getDictations( const std::string& study, std::list<ReportFileInfo>& files /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets the report files for a study.
		 * List the file names of all Basic text, Comprehensive and Enhanced Structured Report objects in the study.
		 * The report objects created by our system will also be returned.
		 *
		 * Used by the report manager to load the external reports - if any.
		 *
		 * @param study [in] the study instance UID
		 * @param files [out] output parameter for the file name of the report objects.
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual void getExternalReports( const std::string& study, std::list<ReportFileInfo>& files /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets selected report files for a study.
		 * The report files will be sorted by creation time in ascending order.
		 *
		 * Used by the report manager to load the reports files (for purge).
		 *
		 * @param study [in] the study instance UID
		 * @param sopiuids [in] the SOP instance UID of the reports to load
		 * @param files [out] output parameter for the file name of the report objects.
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual void getReportsByUid( const std::string& study, const std::list<std::string>& sopiuids,
				std::list<ReportFileInfo>& files /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets all objects (including the PbR) for a study. Series information is also returned for each object.
		 * The result will be sorted by series UID in ascending order.
		 *
		 * Used by MCS servers to collect studies/series from the PACS
		 *
		 * @param study [in] the study instance UID
		 * @param objects [out] output parameter to return the objects in the study
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual void getObjectsByStudy( const std::string& study, std::vector<SeriesObject>& objects /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets the objects having dose information for a study.
		 * Returns the file names of Enhanced SR and XRay Radiation Dose SR objects in the study.
		 *
		 * Used by the dose info collector.
		 *
		 * @param study [in] the study instance UID
		 * @param files [out] output parameter to return the dose SR objects in the study
		 *
		 * @throw DBException if a DB error occurs
		 */
		virtual void getDoseSrObjects( const std::string& study, std::list<std::string>& files /*out*/) noexcept ( false ) = 0;

		/**
		 * Calculates series number for a new series for a study.
		 *
		 * Used by uploading attachments or new objects.
		 *
		 * @param study [in] the study instance UID
		 * @param seriesNumber [out] output parameter to return the series number for the new series
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the study was found
		 */
		virtual bool calcNewSeriesNumber( const std::string& study, int& seriesNumber /*out*/) noexcept ( false ) = 0;

		/**
		 * Gets image information for an object.
		 *
		 * Used by processing eRAD private presentation state data.
		 *
		 * @param study [in] the study instance UID of the object
		 * @param fname [in] the file name of the object
		 * @param imageInfo [out] output parameter for the collected image information
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the object was found
		 */
		virtual bool getImageInfo( const std::string& study, const std::string& fname, ImageInfo& imageInfo /*out*/)
				noexcept ( false ) = 0;


		virtual bool updateObjCacheEntry( const std::string& study ) noexcept ( false ) = 0;

		/**
		 * Gets object cache information for a particular study.
		 *
		 * @param study [in] the study instance UID
		 * @param recentAge [in] provides an <i>age</i> in seconds. An object cache entry within this <i>age</i> is considered as <i>recent</i>.
		 * @param cacheInfo [out] output parameter for the cache information
		 *
		 * @throw DBException if a DB error occurs
		 *
		 * @return true if and only if the study was found in the object cache
		 */
		virtual bool getCacheInfo( const std::string& study, int recentAge, CacheInfo& cacheInfo /*out*/) noexcept ( false ) = 0;

		/**
		 * Create an Object Cache entry for a study.
		 *
		 * @param study the study instance UID
		 * @param status object cache status (0: OK, 1: being purged, 2: error)
		 * @throw DBException if a DB error occurs
		 * @return true if and only if the study entry was inserted to the Object Cache table
		 */
		virtual bool newObjCacheEntry( const std::string& study, int status ) noexcept ( false ) = 0;

};


/* ********************************************
 *
 * Implementation
 *
 * ********************************************/


/**
 * ObjectDB implementation for MySQL
 */
class ObjectDBSql : public ObjectDB, public PSCache {

	public:

		/**
		 * Constructs a new ObjectDBSql object to prepare and execute MySQL queries
		 *
		 * @param pDB connector to the MySQL database
		 */
		ObjectDBSql( DBConnectorSql& pDB );
		virtual ~ObjectDBSql();

		/**
		 * @copydoc ObjectDB::getStudyFiles(const std::string&,std::list<std::string>&)
		 */
		virtual void getStudyFiles( const std::string& study, std::list<std::string>& files /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getSeriesFiles(const std::string&,const std::string&,std::list<std::string>&)
		 */
		virtual void getSeriesFiles( const std::string& study, const std::string& series, std::list<std::string>& files /*out*/)
				noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getFileByUid(const std::string&,const std::string&,std::string&)
		 */
		virtual bool getFileByUid( const std::string& study, const std::string& sopiuid, std::string& file /*out*/) noexcept ( false )
				override;

		/**
		 * @copydoc ObjectDB::getPbRFile(const std::string&,std::string&)
		 */
		virtual bool getPbRFile( const std::string& study, std::string& file /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getPbRInstance(const std::string&,std::string&,std::string&)
		 */
		virtual bool getPbRInstance( const std::string& study, std::string& series /*out*/, std::string& sopInstance /*out*/)
			noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getDictations(const std::string&,std::list<ReportFileInfo>&)
		 */
		virtual void getDictations( const std::string& study, std::list<ReportFileInfo>& files /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getExternalReports(const std::string&,std::list<ReportFileInfo>&)
		 */
		virtual void getExternalReports( const std::string& study, std::list<ReportFileInfo>& files /*out*/) noexcept ( false )
				override;

		/**
		 * @copydoc ObjectDB::getReportsByUid(const std::string&,const std::list<std::string>&,std::list<ReportFileInfo>&)
		 */
		virtual void getReportsByUid( const std::string& study, const std::list<std::string>& sopiuids,
				std::list<ReportFileInfo>& files /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getObjectsByStudy(const std::string&,std::vector<SeriesObject>&)
		 */
		virtual void getObjectsByStudy( const std::string& study, std::vector<SeriesObject>& objects /*out*/) noexcept ( false )
				override;

		/**
		 * @copydoc ObjectDB::getDoseSrObjects(const std::string&,std::list<std::string>&)
		 */
		virtual void getDoseSrObjects( const std::string& study, std::list<std::string>& files /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::calcNewSeriesNumber(const std::string&,int&)
		 */
		virtual bool calcNewSeriesNumber( const std::string& study, int& seriesNumber /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getImageInfo(const std::string&,const std::string&,ImageInfo&)
		 */
		virtual bool getImageInfo( const std::string& study, const std::string& fname, ImageInfo& imageInfo /*out*/)
				noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::updateObjCacheEntry(const std::string&)
		 */
		virtual bool updateObjCacheEntry( const std::string& study ) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::getCacheInfo(const std::string&,int,CacheInfo&)
		 */
		virtual bool getCacheInfo( const std::string& study, int recentAge, CacheInfo& cacheInfo /*out*/) noexcept ( false ) override;

		/**
		 * @copydoc ObjectDB::newObjCacheEntry(const std::string&,int)
		 */
		virtual bool newObjCacheEntry( const std::string& study, int status ) noexcept ( false ) override;

	private:

		std::string objectTbl;
		std::string objectCacheTbl;

		PsQuery studyObjectsQuery;
		PsQuery seriesObjectsQuery;
		PsQuery fileByUidQuery;
		PsQuery pbrQuery;
		PsQuery pbrInstanceQuery;
		PsQuery reportByUidQuery;
		PsQuery dictationsQuery;
		PsQuery externalReportsQuery;
		PsQuery objectsByStudyQuery;
		PsQuery doseSr;
		PsQuery newSeriesNum;
		PsQuery imageInfoQuery;

		PsQuery updateCacheEntry;
		PsQuery cacheInfoQuery;
		PsQuery newCacheEntry;


};

} // namespace

#endif /* IMUTILS_INCLUDE_IMUTILS_OBJECTDB_H_ */
