#ifndef IMUTILS_INCLUDE_IMUTILS_ROLECONFIG_H_
#define IMUTILS_INCLUDE_IMUTILS_ROLECONFIG_H_

#include <string>
#include <exception>
#include <memory>
#include <vector>
#include <map>

#include <grpcpp/grpcpp.h>

namespace erad {
namespace farm {

/**
 * Enumeration of roles in a server farm.
 *
 */
/*
 * XXX: This list needs to be maintained both here, in the serverrole component, and the java side whenever a new role is introduced
 */
enum class Role {
	Application, /**< Application */
	Balancer, /**< Balancer */
	Database, /**< Database */
	Forward, /**< Forward */
	Media, /**< Media */
	Registration,/**< Registration */
	Stream, /**< Stream */
	Web, /**< Web */
	// Helper constant to detect the count of roles.
	// Should never be used as a Role
	// Make sure this is the last value
	UNKNOWN /**< UNKNOWN */
};

/**
 * Gets the Role for a string representation.
 * The string representation expected to be the all lowercase name of the role.
 *
 * @param str the string representation of the role
 * @return the role or Role::UNKNOWN
 */
Role strToRole( const std::string& str );

/**
 * Gets the string representation of role. The string representation is the lowercase name of the role.
 *
 * @param role the role
 * @return the string representation of role
 */
const char* roleToStr( Role role );

/**
 * Custom exception for role configuration related errors
 */
class RoleException : public std::exception {
	public:
		RoleException( const std::string& message ) noexcept;
		virtual ~RoleException();

		virtual const char* what() const noexcept {
			return message.c_str();
		}

	private:
		std::string message;

};

class RoleConfig;

/**
 * Holds informations about the servers in the farm providing a particular role.
 * Can be used to access the IntraCom service at a particular role.
 */
class ServiceInfo {
		friend class RoleConfig;

	protected:
		ServiceInfo();
		ServiceInfo( const ServiceInfo& orig ) = delete;
		ServiceInfo( ServiceInfo&& orig ) = delete;
		~ServiceInfo();

		ServiceInfo& operator=( const ServiceInfo& ) = delete;
		ServiceInfo& operator=( ServiceInfo&& ) = delete;

	public:
		/**
		 * Gets all servers providing the particular role in the farm.
		 * The servers will be returned in the @code{host}:@code{port} format,
		 * where @code{host} is the host name or IP of the server,
		 * and @code{port} is the IntraCom service port.
		 *
		 * @return all servers providing the particular role in the farm
		 */
		const std::vector<std::string>& getAllServers() const {
			return allServers;
		}

		/**
		 * Gets a server to access the IntraCom service at any server providing the particular role in the farm.
		 * Returns the single server providing the role, or the load balancer, if multiple servers provide the role.
		 *
		 * @return a server to access the IntraCom service for the role in the farm
		 */
		const std::string& getAnyServer() const {
			return anyServer;
		}

		/**
		 * Checks whether the server returned by @code{getAnyServer()} is a load balancer or not.
		 *
		 * @return true if and only if the server returned by @code{getAnyServer()} is a load balancer
		 */
		bool isAnyBalancer() const {
			return anyBalancer;
		}

	protected:

		/// all servers providing the role in the farm
		std::vector<std::string> allServers;
		/// a server to access a single server (any server) providing the role
		std::string anyServer;
		/// indicates whether the anyServer is a load balancer
		bool anyBalancer;
};

/**
 * Contains information about the servers in the farm based on the content of the roleconfig.xml
 */
class RoleConfig {
	public:

		/**
		 * Creates a new RoleConfig object reading the servers official role configuration
		 */
		RoleConfig();

		RoleConfig( int port );

		/**
		 * Creates a new RoleConfig object reading a custom role configuration.
		 * Intended for testing purposes only
		 *
		 * @param configFile path the to role configuration
		 */
		RoleConfig( const std::string& configFile );

		RoleConfig( const RoleConfig& orig ) = delete;
		RoleConfig( RoleConfig&& orig ) = delete;

		~RoleConfig();

		RoleConfig& operator=( const RoleConfig& ) = delete;
		RoleConfig& operator=( RoleConfig&& ) = delete;

		/**
		 * Loads the role configuration.
		 *
		 * @throw RoleException if an error occurs
		 */
		void load() noexcept ( false );

		/**
		 * Gets information about the servers providing a particular role in the farm to access the IntraCom service for the role.
		 * Loads the farm configuration if not yet loaded.
		 * Note, the IntraCom service is not available at certain roles, for example at the load balancer.
		 *
		 * @param role the role
		 * @throw RoleException if an error occurs while loading the role configuration or if getting information about the particular role is not supported.
		 * @return information to access the IntraCom service for the role
		 */
		const ServiceInfo& getServiceInfoFor( Role role ) noexcept ( false );

	protected:

		/// path to the role configuration file
		std::string configFile;
		/// The IntraCom service port configured at the farm
		int port = -1;
		/// service information cached for the roles
		ServiceInfo services[ static_cast<int>( Role::UNKNOWN ) + 1 ];
		/// indicates whether the configuration is already loaded
		bool loaded;
};

}
} // namespace erad::farm

#endif /* IMUTILS_INCLUDE_IMUTILS_ROLECONFIG_H_ */
