#ifndef CLOUDDICOMSTORE_H
#define CLOUDDICOMSTORE_H

#include <string>
#include <vector>
#include <Poco/DOM/Document.h>
#include <Poco/DOM/DOMParser.h>
#include <Poco/DOM/NodeList.h>

#include "utils/Http.h"

/**
 * This is the class responsible for communicating with cloud store
 */
class CloudDicomStore {
		/*
		 * HTTP Client object
		 */
		HttpsClient m_HttpsClient;
		/*
		 * HTTP endpoint
		 */
		HttpEndpoint m_healthcare_googleapis_com;
		/*
		 * Base URL of the Dicom store
		 */
		std::string m_BaseUrl;
		/*
		 * token
		 */
		const std::string m_AccessToken;
		/**
		 * Get meta data in JSON format of the response
		 * @param queryUrl dicom object query url
		 * @return JSON string of the query results
		 */
		std::string getMetadataJson( const std::string& queryUrl,
				const std::vector<std::string>& tagsToGet );

	public:

		/**
		 * Constructor for CloudDicomStore
		 * @param dicomStorePath - Dicom store path
		 * @param accessToken    - token
		 * @return nil
		 */
		CloudDicomStore( const std::string& dicomStorePath,
				const std::string& accessToken );

		/**
		 * Get study meta data in JSON format of the response
		 * @param studyUID  - study instance UID
		 * @param tagsToGet - tag list
		 * @return JSON string of the query results
		 */
		std::string getStudyMetadataJson( const std::string& studyUID,
				const std::vector<std::string>& tagsToGet );
		/**
		 * Get Instance meta data in JSON format of the response
		 * @param studyUID    - Study instance UID
		 * @param instanceUID - Sop instance UID
		 * @param tagsToGet   - tag list
		 * @return JSON string of the query results
		 */
		std::string getInstanceMetadataJson( const std::string& studyUID, const std::string& instanceUID,
				const std::vector<std::string>& tagsToGet );

		/**
		 * Function to retrieve the instance from GCP store
		 * @param studyUID       - study instance UID
		 * @param seriesUID      - series instance UID
		 * @param sopInstanceUID - SOP instance UID
		 * @param dicomFilePath  - Retrieved Dicom file path
		 * @return nil
		 */
		void retrieveInstance( const std::string& studyUID, const std::string& seriesUID, const std::string& sopInstanceUID,
				const std::string& dicomFilePath );

		/**
		 * Function to Upload the instance to GCP store
		 * @param filePath          - Dicom file path
		 * @param studyInstanceUID  - study instance UID
		 * @param seriesInstanceUID - series instance UID
		 * @param instanceUID       - SOP instance UID
		 * @return true if success else false
		 */
		bool uploadDicomFile( const std::string& filePath, const std::string& studyInstanceUID,
				const std::string& seriesInstanceUID, const std::string& instanceUID );

		/**
		 * Function to Upload the instance to GCP store
		 * @param filePath          - Dicom file path
		 * @param studyInstanceUID  - study instance UID
		 * @param seriesInstanceUID - series instance UID
		 * @param instanceUID       - SOP instance UID
		 * @return true if success else false
		 */
		bool deleteInstance( const std::string& studyInstanceUID, const std::string& seriesInstanceUID,
				const std::string& instanceUID );

		/**
		 * Function to check the FailureReason and FailureDetail in the response
		 * @param response A pointer to a Poco::XML::Element that represents the root XML element
		 * @returns        true if failureReason is 272 and failureDetail is "already exists".
		 */
		bool checkFileAlreadyExist( const std::string& response );

		/**
		 * Function to find FailedSOPSequence from XML.
		 * @param pRoot          A pointer to a Poco::XML::Element that represents the root XML element
		 * @param failureReason  A reference to an integer where the extracted failure reason will be stored
		 * @param failureDetail  A reference to a string where the extracted failure detail will be stored
		 * @returns              true if success
		 */
		bool findFailedSOPSequence( Poco::XML::Element* pRoot, int& failureReason, std::string& failureDetail );

		/**
		 * Function to extract failure information from XML
		 *
		 * @param pItem          A pointer to a Poco::XML::Element that represents the XML element
		 * @param failureReason  A reference to an integer where the extracted failure reason will be stored
		 * @param failureDetail  A reference to a string where the extracted failure detail will be stored
		 */
		void extractFailureInfo( Poco::XML::Element* pItem, int& failureReason, std::string& failureDetail );

		/**
		 * Function to extract value from XML
		 *
		 * @param pItem          A pointer to a Poco::XML::Element that represents the XML element
		 * @param failureReason  A reference to an integer where the extracted failure reason will be stored
		 * @param failureDetail  A reference to a string where the extracted failure detail will be stored
		 */
		void extractValue( Poco::XML::Element* pInnerDicomAttribute, const std::string& innerKeyword,
			int& failureReason, std::string& failureDetail );
};

#endif // CLOUDDICOMSTORE_H