#ifndef HTTPCLIENTT_H
#define HTTPCLIENTT_H

#include <chrono>
#include <map>
#include <sstream>
#include <stdexcept>
#include <vector>

#include <boost/asio.hpp>
#include <boost/beast.hpp>


struct HttpResponse {
    boost::beast::http::status status;
    std::string body;
};

class HttpTimeoutException : public std::runtime_error
{
public:
    using std::runtime_error::runtime_error;
};


class HttpEndpoint
{
    std::string m_Hostname;
    std::vector<boost::asio::ip::tcp::endpoint> m_Endpoints;

public:
    HttpEndpoint()
    {
    }

    HttpEndpoint(const std::string& hostname, const std::vector<boost::asio::ip::tcp::endpoint>& endpoints)
        : m_Hostname(hostname)
        , m_Endpoints(endpoints)
    {
    }

    const std::string& getHostname() const {
        return m_Hostname;
    }

    const std::vector<boost::asio::ip::tcp::endpoint>& getEndpoints() const {
        return m_Endpoints;
    }
};


template <class t_connector>
class HttpClientT : public t_connector
{
protected:
    boost::asio::io_context m_IOContext;

    std::chrono::milliseconds m_ResolveTimeout;
    std::chrono::milliseconds m_WriteTimeout;
    std::chrono::milliseconds m_ReadTimeout;

protected:
    template <class t_request>
    void sendRequest(typename t_connector::Stream& stream, const t_request& request) {
        bool bSuccess = false;
        boost::system::error_code writeError;

        // Start timer
        boost::asio::steady_timer timer(m_IOContext);
        timer.expires_after(m_WriteTimeout);
        timer.async_wait(
            [&bSuccess, &stream]
                (const boost::system::error_code& error)
            {
                if (!error) {
                    bSuccess = false;
                    boost::beast::get_lowest_layer(stream).cancel();
                }
            }
        );

        // Write request
        boost::beast::http::async_write(stream, request,
            [&bSuccess, &timer, &writeError]
                (const boost::system::error_code& error, std::size_t bytesTransferred)
            {
                writeError = error;

                if (!error) {
                    bSuccess = true;
                    timer.cancel();
                }
            }
        );

        m_IOContext.restart();
        m_IOContext.run();

        if (!bSuccess) {
            const std::string remoteEndpoint = boost::beast::get_lowest_layer(stream).socket().remote_endpoint().address().to_string() + ":" + std::to_string(boost::beast::get_lowest_layer(stream).socket().remote_endpoint().port());

            if (writeError == boost::system::errc::operation_canceled) {
                throw HttpTimeoutException("Writing request to " + remoteEndpoint + " timed out");
            }
            else {
                throw std::runtime_error("Error during writing request to " + remoteEndpoint + ": " + writeError.message());
            }
        }
    }

		template<class t_buffer, class t_response>
		void recvResponse( typename t_connector::Stream& stream, t_buffer& buffer, t_response& response ) {
			bool bSuccess = false;
			boost::system::error_code readError;

			// Start timer
			boost::asio::steady_timer timer( m_IOContext );
			timer.expires_after( m_ReadTimeout );
			timer.async_wait(
					[ &bSuccess, &stream ]
					( const boost::system::error_code& error )
							{
						if ( !error ) {
							bSuccess = false;
							boost::beast::get_lowest_layer( stream ).cancel();
						}
					}
			);

			boost::beast::http::response_parser < boost::beast::http::dynamic_body > response_parser;
			// Disable body size limit (allow for large responses)
			response_parser.body_limit( std::numeric_limits < std::size_t > ::max() );  // No limit

					// Recv request
			boost::beast::http::async_read( stream, buffer, response_parser,
					[ &response, &response_parser, &bSuccess, &timer, &readError ]
					( const boost::system::error_code& error, std::size_t bytesTransferred )
							{
						readError = error;
						response = response_parser.get();
						if ( !error ) {
							bSuccess = true;
							timer.cancel();
						}
					}
			);

			m_IOContext.restart();
			m_IOContext.run();

			if ( !bSuccess ) {
				const std::string remoteEndpoint = boost::beast::get_lowest_layer( stream ).socket().remote_endpoint().address().to_string()
						+ ":" + std::to_string( boost::beast::get_lowest_layer( stream ).socket().remote_endpoint().port() );

				if ( readError == boost::system::errc::operation_canceled ) {
					throw HttpTimeoutException( "Reading response from " + remoteEndpoint + " timed out" );
				}
				else {
					throw std::runtime_error( "Error during reading response from " + remoteEndpoint + ": " + readError.message() );
				}
			}
		}

		template<class t_buffer>
		void recvResponseToFile( typename t_connector::Stream& stream, t_buffer& buffer, const std::string& filePath ) {
			bool bSuccess = false;
			boost::system::error_code readError;

			// Start timer
			boost::asio::steady_timer timer( m_IOContext );
			timer.expires_after( m_ReadTimeout );
			timer.async_wait(
					[ &bSuccess, &stream ]
					( const boost::system::error_code& error )
							{
						if ( !error ) {
							bSuccess = false;
							boost::beast::get_lowest_layer( stream ).cancel();
						}
					}
			);

			boost::beast::error_code ec;
			boost::beast::http::file_body::value_type file;
			file.open( filePath.c_str(), boost::beast::file_mode::write, ec );
			if ( ec ) {
				throw std::runtime_error( "File open error : " + ec.message() );
			}

			boost::beast::http::response_parser < boost::beast::http::file_body > response_parser;
			// Disable body size limit (allow for large responses)
			response_parser.body_limit( std::numeric_limits < std::size_t > ::max() );  // No limit

					// Assign the file to response body
			response_parser.get().body() = std::move( file );

			// Recv request
			boost::beast::http::async_read( stream, buffer, response_parser,
					[ &response_parser, &bSuccess, &timer, &readError ]
					( const boost::system::error_code& error, std::size_t bytesTransferred )
							{
						readError = error;
						if ( !error ) {
							bSuccess = true;
							timer.cancel();
						}
					}
			);

			m_IOContext.restart();
			m_IOContext.run();

			if ( !bSuccess ) {
				const std::string remoteEndpoint = boost::beast::get_lowest_layer( stream ).socket().remote_endpoint().address().to_string()
						+ ":" + std::to_string( boost::beast::get_lowest_layer( stream ).socket().remote_endpoint().port() );

				if ( readError == boost::system::errc::operation_canceled ) {
					throw HttpTimeoutException( "Reading response from " + remoteEndpoint + " timed out" );
				}
				else {
					throw std::runtime_error( "Error during reading response from " + remoteEndpoint + ": " + readError.message() );
				}
			}
		}

public:
    HttpClientT()
        : m_ResolveTimeout(60000)
        , m_WriteTimeout(2000)
        , m_ReadTimeout(60000)
    {
    }

    void setResolveTimeout(const std::chrono::milliseconds& timeout) {
        m_ResolveTimeout = timeout;
    }

    void setWriteTimeout(const std::chrono::milliseconds& timeout) {
        m_ReadTimeout = timeout;
    }

    void setReadTimeout(const std::chrono::milliseconds& timeout) {
        m_ReadTimeout = timeout;
    }

public:
    HttpEndpoint resolve(const std::string& host, const std::string& service) {
        bool bSuccess = false;
        boost::system::error_code resolveError;
        boost::asio::ip::tcp::resolver::results_type resolveResults;

        boost::asio::ip::tcp::resolver resolver(m_IOContext);

        // Start timer
        boost::asio::steady_timer timer(m_IOContext);
        timer.expires_after(m_ResolveTimeout);
        timer.async_wait(
            [&bSuccess, &resolver]
                (const boost::system::error_code& error)
            {
                if (!error) {
                    bSuccess = false;
                    resolver.cancel();
                }
            }
        );

        // Start resolver
        resolver.async_resolve(host, service,
            [&bSuccess, &timer, &resolveError, &resolveResults]
                (const boost::system::error_code& error, const boost::asio::ip::tcp::resolver::results_type& results)
            {
                resolveError = error;

                if (!error) {
                    resolveResults = results;

                    bSuccess = true;
                    timer.cancel();
                }
            }
        );

        m_IOContext.restart();
        m_IOContext.run();

        if (!bSuccess) {
            if (resolveError == boost::system::errc::operation_canceled) {
                throw HttpTimeoutException("Resolving host \"" + host + "\" with service \"" + service + "\" timed out");
            }
            else {
                throw std::runtime_error("Error during resolving host \"" + host + "\" with service \"" + service + "\": " + resolveError.message());
            }
        }

        std::vector<boost::asio::ip::tcp::endpoint> resultEndpoints;
        for (const auto& ep : resolveResults)
            resultEndpoints.push_back(ep);
        
        return HttpEndpoint(host, resultEndpoints);
    }

    HttpEndpoint resolve(const std::string& host) {
        return resolve(host, t_connector::s_Service);
    }

public:

    HttpResponse doRequest(const HttpEndpoint& endpoint, boost::beast::http::verb requestType, const std::string& path, 
                           const std::string& body, const std::map<std::string, std::string>& headers) {
        typename t_connector::Stream stream = this->connect(m_IOContext, endpoint);
        

        // Send request
        if((requestType == boost::beast::http::verb::get) || (requestType == boost::beast::http::verb::post)){
            boost::beast::http::request<boost::beast::http::string_body> request(requestType, path, 11);

            request.set(boost::beast::http::field::host, endpoint.getHostname());
            request.set(boost::beast::http::field::user_agent, "eRAD-PACS");

            for (const auto& header : headers) {
                request.set(header.first, header.second);
            }

            if( !body.empty() ){
                request.body() = body;
                request.prepare_payload();
            }

            sendRequest(stream, request);
        }
        else if( requestType == boost::beast::http::verb::delete_){
            boost::beast::http::request<boost::beast::http::empty_body> request(boost::beast::http::verb::delete_, path, 11);

            request.set(boost::beast::http::field::host, endpoint.getHostname());
            request.set(boost::beast::http::field::user_agent, "eRAD-PACS");

            for (const auto& header : headers) {
                request.set(header.first, header.second);
            }

            sendRequest(stream, request);
        }
        else{
            throw std::runtime_error( "Unsupported request type" );
        }

        // Receive response
        boost::beast::flat_buffer buffer;
        boost::beast::http::response<boost::beast::http::dynamic_body> response;

        recvResponse(stream, buffer, response);

        // Close socket
        this->close(stream);

        // Create an HttpResponse object
        HttpResponse httpResponse;
        httpResponse.status = response.result();
        httpResponse.body = boost::beast::buffers_to_string(response.body().cdata());

        return httpResponse;
    }

    HttpResponse get(const HttpEndpoint& endpoint, const std::string& path, const std::map<std::string, std::string>& headers) {
        return doRequest(endpoint, boost::beast::http::verb::get, path, "", headers);
    }

    std::string get(const HttpEndpoint& endpoint, const std::string& path) {
        std::map<std::string, std::string> headers;

        return get(endpoint, path, headers);
    }

    void getToFile(const HttpEndpoint& endpoint, const std::string& path, const std::map<std::string, std::string>& headers, const std::string& filePath) {
        typename t_connector::Stream stream = this->connect(m_IOContext, endpoint);

        // Send request
        boost::beast::http::request<boost::beast::http::string_body> request(boost::beast::http::verb::get, path, 11);
        request.set(boost::beast::http::field::host, endpoint.getHostname());
        request.set(boost::beast::http::field::user_agent, "eRAD-PACS");

        for (auto it = headers.begin(); it != headers.end(); ++it)
            request.set(it->first, it->second);


        sendRequest(stream, request);

        // Receive response
        boost::beast::flat_buffer buffer;
        recvResponseToFile(stream, buffer, filePath);

        // Close socket
        this->close(stream);
    }

  
    HttpResponse post(const HttpEndpoint& endpoint, const std::string& path, const std::string& body, 
                      const std::map<std::string, std::string>& headers)
    {
        return doRequest(endpoint, boost::beast::http::verb::post, path, body, headers);
    }

    std::string post(const HttpEndpoint& endpoint, const std::string& path, const std::string& body) {
        std::map<std::string, std::string> headers;

        return post(endpoint, path, body, headers);
    }

    HttpResponse deleteFile(const HttpEndpoint& endpoint, const std::string& path, const std::map<std::string, std::string>& headers) {       
        return doRequest(endpoint, boost::beast::http::verb::delete_, path, "", headers);
    }

    std::string deleteFile(const HttpEndpoint& endpoint, const std::string& path) {
        std::map<std::string, std::string> headers;

        return deleteFile(endpoint, path, headers);
    }
};

#endif // HTTPCLIENTT_H