/*
 * Component:  utils
 *
 * Purpose:
 *    Logging implementation using syslog
 *
 */
#ifndef IMDB_LOGGER_H_INCLUDED
#define IMDB_LOGGER_H_INCLUDED

#include <stdarg.h>
#include <string>
#include <map>
#include <vector>

#include "imutils/IMLogger.h"
#include "imutils/StudyDB.h"


namespace erad::log {


/**
 * Encodes string values to be added to oper log entries
 *
 * @param value  the value to be encoded
 * @return the log-encoded value
 */
std::string logValue( const char* value );

/**
 * Helper class to collect study level values to be added to oper log entries from database
 */
class StudyLogInfo : protected erad::db::StudyInfo {
	private:

		erad::db::DBConnector& db;

	public:

		/**
		 * Constructor for creating a new Study Log Info object
		 * @param db database connection
		 * @param studyUID  study instance UID
		 */
		StudyLogInfo( erad::db::DBConnector& db, const char* studyUID );

		/**
		 * Destructor to destroy this Study Log Info object
		 */
		~StudyLogInfo();

		/**
		 * @return  Log-encoded patient name
		 */
		std::string getLogPatientName() const {
			return logValue( patientName.c_str() );
		}

		/**
		 * @return  Log-encoded accession number
		 */
		std::string getLogAccessionNumber() const {
			return logValue( accessionNumber.c_str() );
		}

		/**
		 * @return  Log-encoded modality
		 */
		std::string getLogModality() const {
			return logValue( modality.c_str() );
		}

		/**
		 * @return  Log-encoded patient ID
		 */
		std::string getLogPatientID() const {
			return logValue( patientID.c_str() );
		}

		/**
		 * @return  Log-encoded study date
		 */
		std::string getLogStudyDate() const {
			return logValue( studyDate.c_str() );
		}

		/**
		 * @return  Log-encoded institution
		 */
		std::string getLogInstitution() const {
			return logValue( institution.c_str() );
		}

		/**
		 * @return  Log-encoded study description
		 */
		std::string getLogStudyDescription() const {
			return logValue( studyDescription.c_str() );
		}

		/**
		 * @return  Log-encoded body part
		 */
		std::string getLogBodypart() const {
			return logValue( bodypart.c_str() );
		}

		/**
		 * @return  Log-encoded study instance UID
		 */
		std::string getLogStudyUID() const {
			return logValue( study.c_str() );
		}

		/**
		 * Fills params with study information to be added to a STUDY/OBJECT level oper log entry.
		 * @param params  output parameter to add study log information to
		 */
		void fillParams( std::map<std::string, std::string>& params ) const;
};


/**
 * Wrapper around a log field
 */
class LogField {
		/** Database field name for fields to be added to database log entries as well */
		std::string m_dbname;
		/** Value to be logged */
		std::string m_value;

	public:

		/**
		 * Default constructor creating an empty log field
		 */
		LogField();

		/**
		 * Copy constructor to create a new log field by copying data from another log field object
		 * @param logField original log gield to be copied
		 */
		LogField( const LogField& logField );

		/**
		 * Destructor to destroy the log field
		 */
		~LogField();

		/**
		 * Assignment operator to copy data from another log field object to this log field object
		 * @param logField
		 * @return reference to this log field object
		 */
		LogField& operator=( const LogField& logField );

		/**
		 * @return Database field name of the field
		 */
		inline const std::string& getDbname() const {
			return m_dbname;
		}

		/**
		 * @return  value to be logged
		 */
		inline const std::string& getValue() const {
			return m_value;
		}

		/**
		 * Sets database field name
		 * @param dbname new database field name of the field (empty for fields having no corresponding database field)
		 */
		inline void setDbname( const std::string& dbname ) {
			m_dbname = dbname;
		}

		/**
		 * @param value new value for this log field
		 */
		inline void setValue( const std::string& value ) {
			m_value = value;
		}
};


/**
 * Class for creating oper_info and oper_error log entries and also log them to the log database table
 */
class IMDBLogger : public IMLogger {

		/// Database connection
		erad::db::DBConnector& db;

	public:

		/**
		 * Create new Database Logger object with logging tag.
		 * If no database connection is provided, one will be created for the logging
		 *
		 * @param db database connection
		 * @param tag  Logger tag to be used in log messages
		 */
		IMDBLogger( erad::db::DBConnector& db, const char* tag );

		~IMDBLogger();

		/**
		 * Create log entry in oper_info.log
		 * Log entry will be parsed and added to the log database if needed
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void operInfoLog( const char* mask, ... );

		/**
		 * Create log entry in oper_error.log
		 * Log entry will be parsed and added to the log database if needed
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void operErrorLog( const char* mask, ... );

		/**
		 * Create an oper_info or oper_error log entry using the provided fields.
		 * A log database entry will also be created if the entry intended to be
		 * logged to the database (ie. the level is neither OBJECT nor SERIES).
		 *
		 * fields should contain empty log fields at positions where no log value is specified.
		 *
		 * @param mode    log mode (INFO/ERROR)
		 * @param level   log level (STUDY/SERIES/OBJECT/ADMIN/...)
		 * @param action  log action
		 * @param fields  fields to be logged
		 */
		void operLog( LOG_MODE mode, const char* level, const char* action, const std::vector<LogField>& fields );

		/**
		 * Create an oper_info or oper_error log entry using the provided parameters.
		 * A log database entry will also be created if the entry intended to be
		 * logged to the database (ie. the level is neither OBJECT nor SERIES).
		 *
		 * Parameters intended to be a map of <code>log fields</code> to <code>log values</code>.
		 *
		 * Log field can be either<UL>
		 * <LI> an SQL name of a log database field (only a subset is supported)</LI>
		 * <LI> a number: index of the field in the log file (index starts at 3 - USERID)</LI>
		 * </UL>
		 *
		 * Log value is the value to be logged for the specified field (unescaped)
		 *
		 * @param mode    log mode (INFO/ERROR)
		 * @param level   log level (STUDY/SERIES/OBJECT/ADMIN/...)
		 * @param action  log action
		 * @param params  paremeters holding values to be logged
		 */
		void operLog( LOG_MODE mode, const char* level, const char* action, const std::map<std::string, std::string>& params );

	private:
		/**
		 * Log into database
		 *
		 * @param mask		Log message mask (see printf)
		 * @param args		Arguments to be replaced in mask
		 */
		void dbLog( const char* mask, va_list args );

		/**
		 * Converts log parameters to a vector of log fields
		 * @param params  paremeters holding values to be logged
		 * @param fields  output parameter - the list of log fields generated from log parameters
		 */
		void params2vector( const std::map<std::string, std::string>& params, std::vector<LogField>& fields );

		/**
		 * Adds the generated log entry to the oper_info/oper_error log file.
		 * Intended to be called by public operLog() methods only
		 * @param mode   log mode (INFO/ERROR)
		 * @param mask   log message mask (see printf)
		 */
		void operLog( LOG_MODE mode, const char* mask, ... );

		/**
		 * Checks if callback is registered for the given log level/log action and starts a Callback Hook task
		 * if needed.
		 *
		 * @param level   log level (STUDY/SERIES/OBJECT/ADMIN/...)
		 * @param action  log action
		 */
		void startCallbackTask( const char* level, const char* action );
};

} // namespace erad::log

#endif // IM_LOGGER_H_INCLUDED
