#ifndef UTILS_INCLUDE_UTILS_METAMGR_H_
#define UTILS_INCLUDE_UTILS_METAMGR_H_


#include <string>
#include <set>
#include <filesystem>
#include <utils/Logger.h>
#include <imutils/DBConnector.h>


namespace erad::utils {

/**
 * Enumerated status values of resources in the meta repository
 */
enum class StatusEnum {
	/** The resource exists */
	OK,
	/** The resource has just been created */
	OK_NEW,
	/** The resource has just been copied from legacy directory */
	OK_COPIED,
	/** The resource exists in tempMeta */
	OK_TEMP,
	/** The resource does not exists */
	MISSING,
	/** An error occurred while accessing the meta repository */
	REPOSITORY_ERROR,
	/** Generic error, including when pre-existing .info direcory could not be copied to the meta repository */
	ERROR
};


/**
 * Wrapper class for status values providing conversion methods
 */
class Status {
	private:
		/// the status value
		StatusEnum status;

		// Default constructor is not available
		Status() = delete;

	public:

		/**
		 * Constructs a new Status object
		 * @param status the status value
		 */
		Status( const StatusEnum &status );

		/**
		 * Returns the enumerated status value
		 * @return the status value
		 */
		inline operator StatusEnum() const {
			return status;
		}

		/**
		 * Assigns a new status value
		 *
		 * @param status the new status value
		 *
		 * @return reference to this object
		 */
		inline Status& operator=( StatusEnum status ) {
			this->status = status;
			return *this;
		}

		/**
		 * Checks if the provided status is the same as this status by comparing the enumerated values in the objects
		 * @param s the status to check
		 * @return true if and only if the enumerated status values are equal in the parameter and this objects.
		 */
		inline constexpr bool operator ==( const Status &s ) const {
			return status == s.status;
		}
		/**
		 * Checks if the provided status is different to the status value of this object by comparing the enumerated values in the objects.
		 * @param s the status to check
		 * @return true if and only if the enumerated status values are different in the parameter and this objects.
		 */
		inline constexpr bool operator !=( const Status &s ) const {
			return status != s.status;
		}
		/**
		 * Checks if the provided enumerated status value is the same as the status value of this object.
		 * @param s the status to check
		 * @return true if and only if the provided status value equals to the status value in this object.
		 */
		inline constexpr bool operator ==( StatusEnum s ) const {
			return status == s;
		}
		/**
		 * Checks if the provided enumerated status value is different to the status value of this object.
		 * @param s the status to check
		 * @return true if and only if the provided status value equals to the status value in this object.
		 */
		inline constexpr bool operator !=( StatusEnum s ) const {
			return status != s;
		}

		/**
		 * Converts the status to a string.
		 *
		 * @return a string representation of the enumerated status value
		 */
		operator std::string() const;

		/**
		 * Checks if the status indicates the resource exists
		 * @return true if and only if the status indicates the resource exists
		 */
		operator bool() const;
};



/**
 * Manages resources in the meta repository
 * Copies legacy resources (.info directories) from the dicom repositories if needed
 */
class MetaManager {

		/// The location of the meta repository
		std::string metaRepository;

		/// The location of the tempMeta repository
		std::string tempMetaRepository;

		/// Path to the medsrv tmp directory
		std::filesystem::path tempDir;

		/// Max number of sleeps when waiting for a lock (temp director), when copying legacy resources
		size_t waitCount = 100;
		/// Max time to sleep (in milleseconds) when waiting for a lock (temp director), when copying legacy resources
		size_t sleepMillis = 10; // max 1 seconds

		/// logger for debug messages
		std::shared_ptr<Logger> debugLogger;
		/// logger for error messages
		std::shared_ptr<Logger> errorLogger;

		std::unique_ptr<std::set<std::string>> tempMetaFiles;

	public:

		/**
		 * Constructs a new MetaManager object
		 */
		MetaManager();

		/**
		 * Destroys the MetaManager object
		 */
		~MetaManager() = default;

		// Copy/Move constructors and assignment operators are not available

		MetaManager( const MetaManager& ) = delete;
		MetaManager( MetaManager&& ) = delete;
		MetaManager& operator=( const MetaManager& ) = delete;
		MetaManager& operator=( MetaManager&& ) = delete;

		/**
		 * Locates a resource (study) in the meta repository.
		 * Also supports returning a resource in the tempMeta repository if the resource in the meta repository does not exist.
		 * When tempMeta is requested, but not yet exists, then also checks ~/tmp/ for legacy files.
		 * Creating a resource in meta moves files from tempMeta and ~/tmp/ to the meta repository, and deletes the resource from tempMeta.
		 *
		 * @param study the resource (study) to locate
		 * @param createIfMissing if true then creates the resource if missing
		 * @param useTemp if the meta repository does not exist, then check and use the location of the tempMeta (only when createIfMissing is false)
		 * @param result [output] parameter to return the full path to the resource in the repository
		 * @param dbConn database connection
		 * @param handleLegacy controls whether to take over legacy data from the DICOM repository (default: true)
		 * @return the resource status
		 */
		Status getMeta( const std::string& study, bool createIfMissing, bool useTemp, std::string& result, erad::db::DBConnector* dbConn, bool handleLegacy = true ) const;

		/**
		 * Deletes the resource from the meta repository.
		 * If there are files to be stored 'permanently' after the meta resource is gone,
		 * then these will be moved to the TempMeta repository.
		 *
		 * @param study the resource (study) to delete
		 * @param dbConn database connection
		 * @return true if the resource could be successfully deleted
		 */
		bool deleteMeta( const std::string& study, erad::db::DBConnector* dbConn );

		/**
		 * Deletes the resource from the tempMeta repository.
		 *
		 * @param study the resource (study) to delete
		 * @param dbConn database connection
		 * @return true if the resource could be successfully deleted
		 */
		bool deleteTemp( const std::string& study, erad::db::DBConnector *dbConn ) const;

		/**
		 * Sets a new logger for the debug messages
		 * @param l the new logger for debug messages
		 */
		inline void setDebugLogger( Logger *l ) {
			this->debugLogger.reset( l );
		}

		/**
		 * Sets a new logger for the error messages
		 * @param l the new logger for error messages
		 */
		inline void setErrorLogger( Logger *l ) {
			this->errorLogger.reset( l );
		}

	private:

		Status copyLegacy( const std::string& study, const std::filesystem::path& infoDir, std::string& result, erad::db::DBConnector* dbConn ) const;

		void copyLegacy( const std::filesystem::path& src, const std::filesystem::path& dst ) const;

		Status restoreFromTempMeta( const std::string& study, const std::string& metaDir, erad::db::DBConnector* dbConn ) const;

		Status saveToTempMeta( const std::string& study );

		void initTempMetaFiles();

		bool moveFile( const std::filesystem::path& src, const std::filesystem::path& dst ) const;

		/**
		 * Creates a default storestate.rec file if there was none in the legacy `.info` directory.
		 *
		 * This method does not use the StoreState class to avoid looping back to meta.
		 *
		 * @param storeStatePath path to the storestate.rec file to create
		 */
		void createDefaultStorestate( const std::filesystem::path& storeStatePath ) const;
};

}

#endif /* UTILS_INCLUDE_UTILS_METAMGR_H_ */

