#!/bin/bash
#
# Finds the latest checkOverload delete/move entry in the log and prints out the approximate age of the repository in days, hours, minutes, etc
#
# Authors:
#      Charles Riley <criley@erad.com>
#      Tamas Nemeth <tnemeth@erad.com>
#

. $HOME/etc/setenv.sh

progName=$( basename "$0" )
debug=""
repository=""
displayNowAge="true"
displayCOAge="true"

#
# Converts time provided in seconds to a sum of  days, hours, minutes and seconds
# and then prints the result to standard output 
#
# $1 - time to display in in seconds
#
function show_time_() {

	local num=$1
	local min=0
	local hour=0
	local day=0

	if (( num > 59 )) ; then

		(( sec = num % 60 ))
		(( num = num / 60 ))

		if (( num > 59 )) ; then

			(( min = num % 60 ))
			(( num = num / 60 ))

			if (( num > 23 )) ; then
				(( hour = num % 24 ))
				(( day = num / 24 ))
			else
				(( hour = num ))
			fi
		else
			(( min = num ))
		fi
	else
		(( sec = num ))
	fi

	echo "$day"d "$hour"h "$min"m "$sec"s
}


#
# Prints debug message
#
# $@ - debug message to print
#
function debug_ () {
	[ -z "$debug" ] && return

	echo "DEBUG: $@"
}

#
# Prints usage information then exit
#
# [$1] - error message
#
function usage_() {

	local errMsg="$1"

	[ -n "$errMsg" ] && echo "$errMsg" && echo

	cat <<-EOF
		Checks CheckOverload log entries to calculate the age of a repository.

		Usage:
		    $progName [-h] [-d] [ -n | -c ] { cache | dicom | processed }

		        -h:      prints this help
		        -d:      prints debug information
		        -n:      prints repository age at now
		        -c:      prints repository age at CheckOverload time
		
		If neither -n nor -c was provided then both age will be printed

	EOF

	exit
}


while getopts "cdhn" OPT ; do
	case $OPT in
		c) displayNowAge="false";;
		d) debug=true;;
		h) usage_ ;;
		n) displayCOAge="false";;
		*) usage_ "Unrecognized option: $OPT"
	esac
done

shift $(( OPTIND - 1 ))
repository="$1"

[ -z "$repository" ] && usage_ "No repository was specified"

[ "$repository" != "cache" -a "$repository" != "dicom" -a "$repository" != "processed" ] && usage_ "Unrecognized repository: $repository"


debug_ "check info log"
timeline=$(grep CheckOverload $LOG/info.log | grep -E "^.*CheckOverload.*deletes .*$repository.*" -A 1 | tail -n 1)

if [ -z "$timeline" ] ; then

	for logf in $( ls -t $LOGS_DATED/week/*info* ) ; do
		debug_ "check $( basename $logf )"
		timeline=$(zcat "$logf" | grep CheckOverload | grep -E "^.*CheckOverload.*deletes .*$repository.*" -A 1 | tail -n 1)
		[ -n "$timeline" ] && break;
	done

	if [ -z "$timeline" ] ; then
		echo "No CheckOverload was found in recent logs, probably repository cleanup was not needed"
		exit 0
	fi 
fi

debug_ "$timeline"

hrd=$( echo $timeline| awk -v year=$(date +"%Y") ' { print $1" "$2" "$3" "year }')
hage=$( echo $timeline| awk ' { print $(NF-3)" "$(NF-2)" "$(NF-1)" "$NF }')


debug_ "Run time: $hrd"
debug_ "Age time: $hage"

runstampsecs=$(date --date="$hrd" +%s)
agestampsecs=$(date --date="$hage" +%s)

now=$( date "+%s" )

agesecs=$(( $runstampsecs - $agestampsecs ))
nagesecs=$(( $now - $agestampsecs ))

debug_ "age at CheckOverload time: " $runstampsecs $agestampsecs $agesecs "[$(show_time_ $agesecs)]"
debug_ "age at now               : " $now $agestampsecs $nagesecs "[$(show_time_ $nagesecs)]"

if [ "$displayCOAge" = "true" ] ; then
	echo -n "$repository age on $(hostname -s) at CheckOverload is "
	show_time_ $agesecs
fi

if [ "$displayNowAge" = "true" ] ; then
	echo -n "$repository age on $(hostname -s) at now is "
	show_time_ $nagesecs
fi


